from __future__ import annotations

from coredis.response._callbacks import ResponseCallback
from coredis.response.types import GeoCoordinates, GeoSearchResult
from coredis.typing import AnyStr, List, Optional, ResponseType, Tuple, Union, ValueT


class GeoSearchCallback(
    ResponseCallback[
        ResponseType,
        ResponseType,
        Union[int, Tuple[Union[AnyStr, GeoSearchResult], ...]],
    ]
):
    def transform(
        self, response: ResponseType, **options: Optional[ValueT]
    ) -> Union[int, Tuple[Union[AnyStr, GeoSearchResult], ...]]:
        if options.get("store") or options.get("storedist"):
            return response

        if not (
            options.get("withdist")
            or options.get("withcoord")
            or options.get("withhash")
        ):
            return tuple(list(response))

        results: List[GeoSearchResult] = []

        for result in response:
            results.append(
                GeoSearchResult(
                    result.pop(0),
                    float(result.pop(0)) if options.get("withdist") else None,
                    result.pop(0) if options.get("withhash") else None,
                    (
                        GeoCoordinates(*map(float, result.pop(0)))
                        if options.get("withcoord")
                        else None
                    ),
                )
            )

        return tuple(results)


class GeoCoordinatessCallback(
    ResponseCallback[ResponseType, ResponseType, Tuple[Optional[GeoCoordinates], ...]]
):
    def transform(
        self, response: ResponseType, **options: Optional[ValueT]
    ) -> Tuple[Optional[GeoCoordinates], ...]:
        return tuple(
            map(
                lambda ll: (
                    GeoCoordinates(float(ll[0]), float(ll[1]))
                    if ll is not None
                    else None
                ),
                response,
            )
        )



from __future__ import annotations

import asyncio
import enum
from typing import TYPE_CHECKING, Any

from coredis._utils import b, nativestr
from coredis.modules.response.types import (
    GraphNode,
    GraphPath,
    GraphQueryResult,
    GraphRelation,
    GraphSlowLogInfo,
)
from coredis.response._callbacks import ResponseCallback
from coredis.typing import (
    AnyStr,
    Dict,
    Generic,
    Literal,
    Optional,
    ResponsePrimitive,
    ResponseType,
    StringT,
    Tuple,
    Union,
    ValueT,
)

if TYPE_CHECKING:
    from coredis.client import Client


class ValueTypes(enum.IntEnum):
    VALUE_UNKNOWN = 0
    VALUE_NULL = 1
    VALUE_STRING = 2
    VALUE_INTEGER = 3
    VALUE_BOOLEAN = 4
    VALUE_DOUBLE = 5
    VALUE_ARRAY = 6
    VALUE_EDGE = 7
    VALUE_NODE = 8
    VALUE_PATH = 9
    VALUE_MAP = 10
    VALUE_POINT = 11


PROCEDURE_CALLS = {
    "labels": "db.labels()",
    "relationships": "db.relationshipTypes()",
    "properties": "db.propertyKeys()",
}

SCALAR_MAPPING = {
    ValueTypes.VALUE_INTEGER: int,
    ValueTypes.VALUE_BOOLEAN: lambda v: b(v) == b"true",
    ValueTypes.VALUE_DOUBLE: float,
    ValueTypes.VALUE_STRING: lambda v: v,
    ValueTypes.VALUE_NULL: lambda _: None,
}


class QueryCallback(
    ResponseCallback[ResponseType, ResponseType, GraphQueryResult[AnyStr]],
    Generic[AnyStr],
):
    properties: Dict[int, StringT]
    relationships: Dict[int, StringT]
    labels: Dict[int, StringT]

    def __init__(self, graph: StringT):
        self.graph = graph
        self.properties = {}
        self.relationships = {}
        self.labels = {}

    async def pre_process(
        self, client: Client[Any], response: ResponseType, **options: Optional[ValueT]
    ) -> None:
        if not len(response) == 3:
            return
        result_set = response[1]
        max_label_id, max_relation_id, max_property_id = -1, -1, -1

        cache = client.callback_storage[self.__class__]
        self.labels = cache.setdefault(f"{self.graph}:labels", {})
        self.relationships = cache.setdefault(f"{self.graph}:relationships", {})
        self.properties = cache.setdefault(f"{self.graph}:properties", {})
        for row in result_set:
            for entity in row:
                max_label_id, max_relation_id, max_property_id = self.fetch_max_ids(
                    entity, max_label_id, max_relation_id, max_property_id
                )
        if any(k != -1 for k in [max_label_id, max_relation_id, max_property_id]):
            self.labels, self.relationships, self.properties = await asyncio.gather(
                self.fetch_mapping(max_label_id, "labels", client),
                self.fetch_mapping(max_relation_id, "relationships", client),
                self.fetch_mapping(max_property_id, "properties", client),
            )

    def fetch_max_ids(
        self, entity: Any, max_label_id: int, max_relation_id: int, max_property_id: int
    ) -> Tuple[int, int, int]:
        result_type = entity[0]
        if result_type == ValueTypes.VALUE_NODE:
            for label_id in entity[1][1]:
                max_label_id = max(max_label_id, label_id)
            for property_id in [k[0] for k in entity[1][2]]:
                max_property_id = max(max_property_id, property_id)
        elif result_type == ValueTypes.VALUE_EDGE:
            max_relation_id = max(max_relation_id, entity[1][1])
            for property_id in [k[0] for k in entity[1][4]]:
                max_property_id = max(max_property_id, property_id)
        elif result_type == ValueTypes.VALUE_PATH:
            for segment in entity[1]:
                max_label_id, max_relation_id, max_property_id = self.fetch_max_ids(
                    segment, max_label_id, max_relation_id, max_property_id
                )
        elif result_type == ValueTypes.VALUE_ARRAY:
            for segment in entity[1]:
                max_label_id, max_relation_id, max_property_id = self.fetch_max_ids(
                    segment, max_label_id, max_relation_id, max_property_id
                )
        return max_label_id, max_relation_id, max_property_id

    async def fetch_mapping(
        self,
        max_id: int,
        type: Literal["labels", "properties", "relationships"],
        client: Client[Any],
    ) -> Dict[int, StringT]:
        cache = client.callback_storage[self.__class__]
        if max_id > max(cache[f"{self.graph}:{type}"] or [-1]):
            cache[f"{self.graph}:{type}"] = dict(
                enumerate(
                    [
                        k[0]
                        for k in (
                            await client.graph.ro_query(
                                self.graph, f"CALL {PROCEDURE_CALLS[type]}"
                            )
                        ).result_set
                    ],
                )
            )
        return cache[f"{self.graph}:{type}"]

    def transform(
        self, response: ResponseType, **options: Optional[ValueT]
    ) -> GraphQueryResult[AnyStr]:
        result_set = []
        headers = []
        if len(response) == 3:
            headers = [k[1] for k in response[0]]
            stats = response[2]
            for result in response[1]:
                entities = []
                for entity in result:
                    entities.append(self.parse_entity(entity))
                result_set.append(entities)
        else:
            stats = response[0]
        stats_mapping = dict(
            map(
                lambda v: int(v) if v.isalnum() else v,
                map(lambda v: v.strip(), nativestr(m).split(":")),
            )
            for m in stats
        )
        return GraphQueryResult(tuple(headers), tuple(result_set), stats_mapping)

    def parse_entity(self, entity):
        result_type = entity[0]
        if result_type in [
            ValueTypes.VALUE_NULL,
            ValueTypes.VALUE_STRING,
            ValueTypes.VALUE_INTEGER,
            ValueTypes.VALUE_BOOLEAN,
            ValueTypes.VALUE_DOUBLE,
        ]:
            return SCALAR_MAPPING[result_type](entity[1])
        elif result_type == ValueTypes.VALUE_MAP:
            it = iter(entity[1])
            return dict(zip(it, map(self.parse_entity, it)))
        elif result_type == ValueTypes.VALUE_ARRAY:
            return [self.parse_entity(k) for k in entity[1]]
        elif result_type == ValueTypes.VALUE_POINT:
            return tuple(map(float, entity[1]))
        elif result_type == ValueTypes.VALUE_EDGE:
            return GraphRelation(
                id=entity[1][0],
                type=self.relationships[entity[1][1]],
                src_node=entity[1][2],
                destination_node=entity[1][3],
                properties=dict(
                    (
                        self.properties[k[0]],
                        self.parse_entity((k[1], k[2])),
                    )
                    for k in entity[1][4]
                ),
            )
        elif result_type == ValueTypes.VALUE_NODE:
            return GraphNode(
                id=entity[1][0],
                labels=set(self.labels[k] for k in entity[1][1]),
                properties=dict(
                    (
                        self.properties[k[0]],
                        self.parse_entity((k[1], k[2])),
                    )
                    for k in entity[1][2]
                ),
            )
        elif result_type == ValueTypes.VALUE_PATH:
            nodes, relations = entity[1]
            nodes = self.parse_entity(nodes)
            relations = self.parse_entity(relations)
            return GraphPath(nodes, relations)


class GraphSlowLogCallback(
    ResponseCallback[
        ResponseType, ResponseType, Union[Tuple[GraphSlowLogInfo, ...], bool]
    ]
):
    def transform(
        self, response: ResponseType, **kwargs: Optional[ValueT]
    ) -> Union[Tuple[GraphSlowLogInfo, ...], bool]:
        return tuple(
            GraphSlowLogInfo(int(k[0]), k[1], k[2], float(k[3])) for k in response
        )


class ConfigGetCallback(
    ResponseCallback[
        ResponseType,
        ResponseType,
        Union[ResponsePrimitive, Dict[AnyStr, ResponsePrimitive]],
    ]
):
    def transform(
        self, response: ResponseType, **options: Optional[ValueT]
    ) -> Union[ResponsePrimitive, Dict[AnyStr, ResponsePrimitive]]:
        if isinstance(response, list):
            if isinstance(response[0], list):
                return dict(response)
            else:
                return response[1]



from __future__ import annotations

import asyncio
import contextlib
import contextvars
import functools
import inspect
import textwrap
from abc import ABCMeta
from ssl import SSLContext
from typing import TYPE_CHECKING, Any, cast, overload

from deprecated.sphinx import versionadded

from coredis._utils import b, hash_slot
from coredis.cache import AbstractCache, SupportsClientTracking
from coredis.client.basic import Client, Redis
from coredis.commands._key_spec import KeySpec
from coredis.commands.constants import CommandName, NodeFlag
from coredis.commands.pubsub import ClusterPubSub, ShardedPubSub
from coredis.connection import RedisSSLContext
from coredis.exceptions import (
    AskError,
    BusyLoadingError,
    ClusterDownError,
    ClusterError,
    ConnectionError,
    MovedError,
    RedisClusterException,
    TimeoutError,
    TryAgainError,
    WatchError,
)
from coredis.globals import MODULE_GROUPS, READONLY_COMMANDS
from coredis.pool import ClusterConnectionPool
from coredis.pool.nodemanager import ManagedNode
from coredis.response._callbacks import AsyncPreProcessingCallback, NoopCallback
from coredis.retry import CompositeRetryPolicy, ConstantRetryPolicy, RetryPolicy
from coredis.typing import (
    AnyStr,
    AsyncIterator,
    Awaitable,
    Callable,
    ContextManager,
    Coroutine,
    Dict,
    Iterable,
    Iterator,
    List,
    Literal,
    Node,
    Optional,
    Parameters,
    ParamSpec,
    ResponseType,
    Set,
    StringT,
    Tuple,
    Type,
    TypeVar,
    ValueT,
)

P = ParamSpec("P")
R = TypeVar("R")

if TYPE_CHECKING:
    import coredis.pipeline


class ClusterMeta(ABCMeta):
    ROUTING_FLAGS: Dict[bytes, NodeFlag]
    SPLIT_FLAGS: Dict[bytes, NodeFlag]
    RESULT_CALLBACKS: Dict[bytes, Callable[..., ResponseType]]
    NODE_FLAG_DOC_MAPPING = {
        NodeFlag.PRIMARIES: "all primaries",
        NodeFlag.REPLICAS: "all replicas",
        NodeFlag.RANDOM: "a random node",
        NodeFlag.ALL: "all nodes",
        NodeFlag.SLOT_ID: "one or more nodes based on the slots provided",
    }

    def __new__(
        cls, name: str, bases: Tuple[type, ...], namespace: Dict[str, object]
    ) -> ClusterMeta:
        kls = super().__new__(cls, name, bases, namespace)
        methods = dict(k for k in inspect.getmembers(kls) if inspect.isfunction(k[1]))
        for module in MODULE_GROUPS:
            methods.update(
                dict(
                    (f"{module.MODULE}.{k[0]}", k[1])
                    for k in inspect.getmembers(module)
                    if inspect.isfunction(k[1])
                )
            )
        for method_name, method in methods.items():
            doc_addition = ""
            cmd = getattr(method, "__coredis_command", None)
            if cmd:
                if cmd.cluster.route:
                    kls.ROUTING_FLAGS[cmd.command] = cmd.cluster.route
                    aggregate_note = ""
                    if cmd.cluster.multi_node:
                        if cmd.cluster.combine:
                            aggregate_note = (
                                f"and return {cmd.cluster.combine.response_policy}"
                            )
                        else:
                            aggregate_note = (
                                "and a mapping of nodes to results will be returned"
                            )
                    doc_addition = f"""
.. admonition:: Cluster note

   The command will be run on **{cls.NODE_FLAG_DOC_MAPPING[cmd.cluster.route]}** {aggregate_note}
                    """
                elif cmd.cluster.split and cmd.cluster.combine:
                    kls.SPLIT_FLAGS[cmd.command] = cmd.cluster.split
                    doc_addition = f"""
.. admonition:: Cluster note

   The command will be run on **{cls.NODE_FLAG_DOC_MAPPING[cmd.cluster.split]}**
   by distributing the keys to the appropriate nodes and return
   {cmd.cluster.combine.response_policy}.

   To disable this behavior set :paramref:`RedisCluster.non_atomic_cross_slot` to ``False``
                """
                if cmd.cluster.multi_node:
                    kls.RESULT_CALLBACKS[cmd.command] = cmd.cluster.combine
            if doc_addition and not hasattr(method, "__cluster_docs"):
                if not getattr(method, "__coredis_module", None):

                    def __w(
                        func: Callable[P, Awaitable[R]]
                    ) -> Callable[P, Awaitable[R]]:
                        @functools.wraps(func)
                        async def _w(*a: P.args, **k: P.kwargs) -> R:
                            return await func(*a, **k)

                        _w.__doc__ = f"""{textwrap.dedent(method.__doc__ or "")}
{doc_addition}
                    """
                        return _w

                    wrapped = __w(method)
                    setattr(wrapped, "__cluster_docs", doc_addition)
                    setattr(kls, method_name, wrapped)
                else:
                    method.__doc__ = f"""{textwrap.dedent(method.__doc__ or "")}
{doc_addition}
                    """
                    setattr(method, "__cluster_docs", doc_addition)
        return kls


RedisClusterT = TypeVar("RedisClusterT", bound="RedisCluster[Any]")
RedisClusterStringT = TypeVar("RedisClusterStringT", bound="RedisCluster[str]")
RedisClusterBytesT = TypeVar("RedisClusterBytesT", bound="RedisCluster[bytes]")


class RedisCluster(
    Client[AnyStr],
    metaclass=ClusterMeta,
):
    MAX_RETRIES = 16
    ROUTING_FLAGS: Dict[bytes, NodeFlag] = {}
    SPLIT_FLAGS: Dict[bytes, NodeFlag] = {}
    RESULT_CALLBACKS: Dict[bytes, Callable[..., Any]] = {}

    connection_pool: ClusterConnectionPool

    @overload
    def __init__(
        self: RedisCluster[bytes],
        host: Optional[str] = ...,
        port: Optional[int] = ...,
        *,
        startup_nodes: Optional[Iterable[Node]] = ...,
        stream_timeout: Optional[float] = ...,
        connect_timeout: Optional[float] = ...,
        ssl: bool = ...,
        ssl_context: Optional[SSLContext] = ...,
        ssl_keyfile: Optional[str] = ...,
        ssl_certfile: Optional[str] = ...,
        ssl_cert_reqs: Optional[Literal["optional", "required", "none"]] = ...,
        ssl_check_hostname: Optional[bool] = ...,
        ssl_ca_certs: Optional[str] = ...,
        max_connections: int = ...,
        max_connections_per_node: bool = ...,
        readonly: bool = ...,
        read_from_replicas: bool = ...,
        reinitialize_steps: Optional[int] = ...,
        skip_full_coverage_check: bool = ...,
        nodemanager_follow_cluster: bool = ...,
        decode_responses: Literal[False] = ...,
        connection_pool: Optional[ClusterConnectionPool] = ...,
        connection_pool_cls: Type[ClusterConnectionPool] = ...,
        protocol_version: Literal[2, 3] = ...,
        verify_version: bool = ...,
        non_atomic_cross_slot: bool = ...,
        cache: Optional[AbstractCache] = ...,
        noreply: bool = ...,
        noevict: bool = ...,
        notouch: bool = ...,
        retry_policy: RetryPolicy = ...,
        **kwargs: Any,
    ) -> None: ...

    @overload
    def __init__(
        self: RedisCluster[str],
        host: Optional[str] = ...,
        port: Optional[int] = ...,
        *,
        startup_nodes: Optional[Iterable[Node]] = ...,
        stream_timeout: Optional[float] = ...,
        connect_timeout: Optional[float] = ...,
        ssl: bool = ...,
        ssl_context: Optional[SSLContext] = ...,
        ssl_keyfile: Optional[str] = ...,
        ssl_certfile: Optional[str] = ...,
        ssl_cert_reqs: Optional[Literal["optional", "required", "none"]] = ...,
        ssl_check_hostname: Optional[bool] = ...,
        ssl_ca_certs: Optional[str] = ...,
        max_connections: int = ...,
        max_connections_per_node: bool = ...,
        readonly: bool = ...,
        read_from_replicas: bool = ...,
        reinitialize_steps: Optional[int] = ...,
        skip_full_coverage_check: bool = ...,
        nodemanager_follow_cluster: bool = ...,
        decode_responses: Literal[True],
        connection_pool: Optional[ClusterConnectionPool] = ...,
        connection_pool_cls: Type[ClusterConnectionPool] = ...,
        protocol_version: Literal[2, 3] = ...,
        verify_version: bool = ...,
        non_atomic_cross_slot: bool = ...,
        cache: Optional[AbstractCache] = ...,
        noreply: bool = ...,
        noevict: bool = ...,
        notouch: bool = ...,
        retry_policy: RetryPolicy = ...,
        **kwargs: Any,
    ) -> None: ...

    def __init__(
        self,
        host: Optional[str] = None,
        port: Optional[int] = None,
        *,
        startup_nodes: Optional[Iterable[Node]] = None,
        stream_timeout: Optional[float] = None,
        connect_timeout: Optional[float] = None,
        ssl: bool = False,
        ssl_context: Optional[SSLContext] = None,
        ssl_keyfile: Optional[str] = None,
        ssl_certfile: Optional[str] = None,
        ssl_cert_reqs: Optional[Literal["optional", "required", "none"]] = None,
        ssl_check_hostname: Optional[bool] = None,
        ssl_ca_certs: Optional[str] = None,
        max_connections: int = 32,
        max_connections_per_node: bool = False,
        readonly: bool = False,
        read_from_replicas: bool = False,
        reinitialize_steps: Optional[int] = None,
        skip_full_coverage_check: bool = False,
        nodemanager_follow_cluster: bool = True,
        decode_responses: bool = False,
        connection_pool: Optional[ClusterConnectionPool] = None,
        connection_pool_cls: Type[ClusterConnectionPool] = ClusterConnectionPool,
        protocol_version: Literal[2, 3] = 3,
        verify_version: bool = True,
        non_atomic_cross_slot: bool = True,
        cache: Optional[AbstractCache] = None,
        noreply: bool = False,
        noevict: bool = False,
        notouch: bool = False,
        retry_policy: RetryPolicy = CompositeRetryPolicy(
            ConstantRetryPolicy((ClusterDownError,), 2, 0.1),
            ConstantRetryPolicy(
                (
                    ConnectionError,
                    TimeoutError,
                ),
                2,
                0.1,
            ),
        ),
        **kwargs: Any,
    ) -> None:
        """

        Changes
          - .. versionadded:: 4.12.0

            - :paramref:`retry_policy`
            - :paramref:`noevict`
            - :paramref:`notouch`
            - :meth:`RedisCluster.ensure_persistence` context manager
            - Redis Module support

              - RedisJSON: :attr:`RedisCluster.json`
              - RedisBloom:

                - BloomFilter: :attr:`RedisCluster.bf`
                - CuckooFilter: :attr:`RedisCluster.cf`
                - CountMinSketch: :attr:`RedisCluster.cms`
                - TopK: :attr:`RedisCluster.topk`
                - TDigest: :attr:`RedisCluster.tdigest`
              - RedisTimeSeries: :attr:`RedisCluster.timeseries`
              - RedisGraph: :attr:`RedisCluster.graph`
              - RediSearch:

                - Search & Aggregation: :attr:`RedisCluster.search`
                - Autocomplete: Added :attr:`RedisCluster.autocomplete`

          - .. versionchanged:: 4.4.0

            - :paramref:`nodemanager_follow_cluster` now defaults to ``True``

          - .. deprecated:: 4.4.0

            - The :paramref:`readonly` argument is deprecated in favour of
              :paramref:`read_from_replicas`

          - .. versionadded:: 4.3.0

            - Added :paramref:`connection_pool_cls`

          - .. versionchanged:: 4.0.0

            - :paramref:`non_atomic_cross_slot` defaults to ``True``
            - :paramref:`protocol_version`` defaults to ``3``

          - .. versionadded:: 3.11.0

            - Added :paramref:`noreply`

          - .. versionadded:: 3.10.0

            - Synchronized ssl constructor parameters with :class:`coredis.Redis`

          - .. versionadded:: 3.9.0

            - If :paramref:`cache` is provided the client will check & populate
              the cache for read only commands and invalidate it for commands
              that could change the key(s) in the request.

          - .. versionadded:: 3.6.0

            - The :paramref:`non_atomic_cross_slot` parameter was added

          - .. versionchanged:: 3.5.0

            - The :paramref:`verify_version` parameter now defaults to ``True``

          - .. versionadded:: 3.1.0

            - The :paramref:`protocol_version` and :paramref:`verify_version`
              parameters were added

        :param host: Can be used to point to a startup node
        :param port: Can be used to point to a startup node
        :param startup_nodes: List of nodes that initial bootstrapping can be done
         from
        :param stream_timeout: Timeout (seconds) when reading responses from the server
        :param connect_timeout: Timeout (seconds) for establishing a connection to the server
        :param ssl: Whether to use an SSL connection
        :param ssl_context: If provided the :class:`ssl.SSLContext` will be used when
         establishing the connection. Otherwise either the default context (if no other
         ssl related parameters are provided) or a custom context based on the other
         ``ssl_*`` parameters will be used.
        :param ssl_keyfile: Path of the private key to use
        :param ssl_certfile: Path to the certificate corresponding to :paramref:`ssl_keyfile`
        :param ssl_cert_reqs: Whether to try to verify the server's certificates and
         how to behave if verification fails (See :attr:`ssl.SSLContext.verify_mode`).
        :param ssl_check_hostname: Whether to enable hostname checking when establishing
         an ssl connection.
        :param ssl_ca_certs: Path to a concatenated certificate authority file or a directory
         containing several CA certifcates to use  for validating the server's certificates
         when :paramref:`ssl_cert_reqs` is not ``"none"``
         (See :meth:`ssl.SSLContext.load_verify_locations`).
        :param max_connections: Maximum number of connections that should be kept open at one time
        :param max_connections_per_node:
        :param read_from_replicas: If ``True`` the client will route readonly commands to replicas
        :param reinitialize_steps: Number of moved errors that result in a cluster
         topology refresh using the startup nodes provided
        :param skip_full_coverage_check: Skips the check of cluster-require-full-coverage config,
         useful for clusters without the CONFIG command (like aws)
        :param nodemanager_follow_cluster: The node manager will during initialization try the
         last set of nodes that it was operating on. This will allow the client to drift along
         side the cluster if the cluster nodes move around alot.
        :param decode_responses: If ``True`` string responses from the server
         will be decoded using :paramref:`encoding` before being returned.
         (See :ref:`handbook/encoding:encoding/decoding`)
        :param connection_pool: The connection pool instance to use. If not provided
         a new pool will be assigned to this client.
        :param connection_pool_cls: The connection pool class to use when constructing
         a connection pool for this instance.
        :param protocol_version: Whether to use the RESP (``2``) or RESP3 (``3``)
         protocol for parsing responses from the server (Default ``3``).
         (See :ref:`handbook/response:redis response`)
        :param verify_version: Validate redis server version against the documented
         version introduced before executing a command and raises a
         :exc:`CommandNotSupportedError` error if the required version is higher than
         the reported server version
        :param non_atomic_cross_slot: If ``True`` certain commands that can operate
         on multiple keys (cross slot) will be split across the relevant nodes by
         mapping the keys to the appropriate slot and the result merged before being
         returned.
        :param cache: If provided the cache will be used to avoid requests for read only
         commands if the client has already requested the data and it hasn't been invalidated.
         The cache is responsible for any mutations to the keys that happen outside of this client
        :param noreply: If ``True`` the client will not request a response for any
         commands sent to the server.
        :param noevict: Ensures that connections from the client will be excluded from the
         client eviction process even if we're above the configured client eviction threshold.
        :param notouch: Ensures that commands sent by the client will not alter the LRU/LFU
         of the keys they access.
        :param retry_policy: The retry policy to use when interacting with the cluster
        """

        if "db" in kwargs:  # noqa
            raise RedisClusterException(
                "Argument 'db' is not possible to use in cluster mode"
            )

        if connection_pool:
            pool = connection_pool
        else:
            startup_nodes = [] if startup_nodes is None else list(startup_nodes)

            # Support host/port as argument

            if host:
                startup_nodes.append(
                    Node(
                        host=host,
                        port=port if port else 7000,
                    )
                )
            if ssl_context is not None:
                kwargs["ssl_context"] = ssl_context
            elif ssl:
                ssl_context = RedisSSLContext(
                    ssl_keyfile,
                    ssl_certfile,
                    ssl_cert_reqs,
                    ssl_ca_certs,
                    ssl_check_hostname,
                ).get()
                kwargs["ssl_context"] = ssl_context

            pool = connection_pool_cls(
                startup_nodes=startup_nodes,
                max_connections=max_connections,
                reinitialize_steps=reinitialize_steps,
                max_connections_per_node=max_connections_per_node,
                skip_full_coverage_check=skip_full_coverage_check,
                nodemanager_follow_cluster=nodemanager_follow_cluster,
                read_from_replicas=readonly or read_from_replicas,
                decode_responses=decode_responses,
                protocol_version=protocol_version,
                noreply=noreply,
                noevict=noevict,
                notouch=notouch,
                stream_timeout=stream_timeout,
                connect_timeout=connect_timeout,
                **kwargs,
            )

        super().__init__(
            stream_timeout=stream_timeout,
            connect_timeout=connect_timeout,
            connection_pool=pool,
            connection_pool_cls=connection_pool_cls,
            decode_responses=decode_responses,
            verify_version=verify_version,
            protocol_version=protocol_version,
            noreply=noreply,
            noevict=noevict,
            notouch=notouch,
            retry_policy=retry_policy,
            **kwargs,
        )

        self.refresh_table_asap: bool = False
        self.route_flags: Dict[bytes, NodeFlag] = self.__class__.ROUTING_FLAGS.copy()
        self.split_flags: Dict[bytes, NodeFlag] = self.__class__.SPLIT_FLAGS.copy()
        self.result_callbacks: Dict[bytes, Callable[..., Any]] = (
            self.__class__.RESULT_CALLBACKS.copy()
        )
        self.non_atomic_cross_slot = non_atomic_cross_slot
        self.cache = cache
        self._decodecontext: contextvars.ContextVar[Optional[bool],] = (
            contextvars.ContextVar("decode", default=None)
        )
        self._encodingcontext: contextvars.ContextVar[Optional[str],] = (
            contextvars.ContextVar("decode", default=None)
        )

    @classmethod
    @overload
    def from_url(
        cls: Type[RedisClusterBytesT],
        url: str,
        *,
        db: Optional[int] = ...,
        skip_full_coverage_check: bool = ...,
        decode_responses: Literal[False] = ...,
        protocol_version: Literal[2, 3] = ...,
        verify_version: bool = ...,
        noreply: bool = ...,
        noevict: bool = ...,
        notouch: bool = ...,
        retry_policy: RetryPolicy = ...,
        **kwargs: Any,
    ) -> RedisClusterBytesT: ...

    @classmethod
    @overload
    def from_url(
        cls: Type[RedisClusterStringT],
        url: str,
        *,
        db: Optional[int] = ...,
        skip_full_coverage_check: bool = ...,
        decode_responses: Literal[True],
        protocol_version: Literal[2, 3] = ...,
        verify_version: bool = ...,
        noreply: bool = ...,
        noevict: bool = ...,
        notouch: bool = ...,
        retry_policy: RetryPolicy = ...,
        **kwargs: Any,
    ) -> RedisClusterStringT: ...

    @classmethod
    def from_url(
        cls: Type[RedisClusterT],
        url: str,
        *,
        db: Optional[int] = None,
        skip_full_coverage_check: bool = False,
        decode_responses: bool = False,
        protocol_version: Literal[2, 3] = 3,
        verify_version: bool = True,
        noreply: bool = False,
        noevict: bool = False,
        notouch: bool = False,
        retry_policy: RetryPolicy = CompositeRetryPolicy(
            ConstantRetryPolicy((ClusterDownError,), 2, 0.1),
            ConstantRetryPolicy(
                (
                    ConnectionError,
                    TimeoutError,
                ),
                2,
                0.1,
            ),
        ),
        **kwargs: Any,
    ) -> RedisClusterT:
        """
        Return a Cluster client object configured from the startup node in URL,
        which must use either the ``redis://`` scheme
        `<http://www.iana.org/assignments/uri-schemes/prov/redis>`_

        For example:

            - ``redis://[:password]@localhost:6379``
            - ``rediss://[:password]@localhost:6379``

        :paramref:`url` and :paramref:`kwargs` are passed as is to
        the :func:`coredis.ConnectionPool.from_url`.
        """
        if decode_responses:
            return cls(
                decode_responses=True,
                protocol_version=protocol_version,
                verify_version=verify_version,
                noreply=noreply,
                retry_policy=retry_policy,
                connection_pool=ClusterConnectionPool.from_url(
                    url,
                    db=db,
                    skip_full_coverage_check=skip_full_coverage_check,
                    decode_responses=decode_responses,
                    protocol_version=protocol_version,
                    noreply=noreply,
                    noevict=noevict,
                    notouch=notouch,
                    **kwargs,
                ),
            )
        else:
            return cls(
                decode_responses=False,
                protocol_version=protocol_version,
                verify_version=verify_version,
                noreply=noreply,
                retry_policy=retry_policy,
                connection_pool=ClusterConnectionPool.from_url(
                    url,
                    db=db,
                    skip_full_coverage_check=skip_full_coverage_check,
                    decode_responses=decode_responses,
                    protocol_version=protocol_version,
                    noreply=noreply,
                    noevict=noevict,
                    notouch=notouch,
                    **kwargs,
                ),
            )

    async def initialize(self) -> RedisCluster[AnyStr]:
        if self.refresh_table_asap:
            self.connection_pool.initialized = False
        await super().initialize()
        if self.cache:
            self.cache = await self.cache.initialize(self)
        self.refresh_table_asap = False
        return self

    def __repr__(self) -> str:
        servers = list(
            {
                "{}:{}".format(info.host, info.port)
                for info in self.connection_pool.nodes.startup_nodes
            }
        )
        servers.sort()

        return "{}<{}>".format(type(self).__name__, ", ".join(servers))

    @property
    def all_nodes(self) -> Iterator[Redis[AnyStr]]:
        """ """
        for node in self.connection_pool.nodes.all_nodes():
            yield cast(
                Redis[AnyStr],
                self.connection_pool.nodes.get_redis_link(node.host, node.port),
            )

    @property
    def primaries(self) -> Iterator[Redis[AnyStr]]:
        """ """
        for primary in self.connection_pool.nodes.all_primaries():
            yield cast(
                Redis[AnyStr],
                self.connection_pool.nodes.get_redis_link(primary.host, primary.port),
            )

    @property
    def replicas(self) -> Iterator[Redis[AnyStr]]:
        """ """
        for replica in self.connection_pool.nodes.all_replicas():
            yield cast(
                Redis[AnyStr],
                self.connection_pool.nodes.get_redis_link(replica.host, replica.port),
            )

    @property
    def num_replicas_per_shard(self) -> int:
        """
        Number of replicas per shard of the cluster determined by
        initial cluster topology discovery
        """
        return self.connection_pool.nodes.replicas_per_shard

    async def _ensure_initialized(self) -> None:
        if not self.connection_pool.initialized or self.refresh_table_asap:
            await self

    def _determine_slots(
        self, command: bytes, *args: ValueT, **options: Optional[ValueT]
    ) -> Set[int]:
        """Determines the slots the command and args would touch"""
        keys = cast(Tuple[ValueT, ...], options.get("keys")) or KeySpec.extract_keys(
            command, *args, readonly_command=self.connection_pool.read_from_replicas
        )
        if (
            command
            in {
                CommandName.EVAL,
                CommandName.EVAL_RO,
                CommandName.EVALSHA,
                CommandName.EVALSHA_RO,
                CommandName.FCALL,
                CommandName.FCALL_RO,
                CommandName.PUBLISH,
            }
            and not keys
        ):
            return set()

        return {hash_slot(b(key)) for key in keys}

    def _merge_result(
        self,
        command: bytes,
        res: Dict[str, R],
        **kwargs: Optional[ValueT],
    ) -> R:
        assert command in self.result_callbacks
        return cast(
            R,
            self.result_callbacks[command](
                res, version=self.protocol_version, **kwargs
            ),
        )

    def determine_node(
        self, command: bytes, **kwargs: Optional[ValueT]
    ) -> Optional[List[ManagedNode]]:
        node_flag = self.route_flags.get(command)
        if command in self.split_flags and self.non_atomic_cross_slot:
            node_flag = self.split_flags[command]

        if node_flag == NodeFlag.RANDOM:
            return [self.connection_pool.nodes.random_node(primary=True)]
        elif node_flag == NodeFlag.PRIMARIES:
            return list(self.connection_pool.nodes.all_primaries())
        elif node_flag == NodeFlag.ALL:
            return list(self.connection_pool.nodes.all_nodes())
        elif node_flag == NodeFlag.SLOT_ID:
            slot_id: Optional[ValueT] = kwargs.get("slot_id")
            node_from_slot = (
                self.connection_pool.nodes.node_from_slot(int(slot_id))
                if slot_id is not None
                else None
            )
            if node_from_slot:
                return [node_from_slot]
        return None

    async def on_connection_error(self, _: BaseException) -> None:
        self.connection_pool.disconnect()
        self.connection_pool.reset()
        self.refresh_table_asap = True

    async def on_cluster_down_error(self, _: BaseException) -> None:
        self.connection_pool.disconnect()
        self.connection_pool.reset()
        self.refresh_table_asap = True

    async def execute_command(
        self,
        command: bytes,
        *args: ValueT,
        callback: Callable[..., R] = NoopCallback(),
        **kwargs: Optional[ValueT],
    ) -> R:
        """
        Sends a command to one or many nodes in the cluster
        with retries based on :paramref:`RedisCluster.retry_policy`
        """

        return await self.retry_policy.call_with_retries(
            lambda: self._execute_command(command, *args, callback=callback, **kwargs),
            failure_hook={
                ConnectionError: self.on_connection_error,
                ClusterDownError: self.on_cluster_down_error,
            },
            before_hook=self._ensure_initialized,
        )

    async def _execute_command(
        self,
        command: bytes,
        *args: ValueT,
        callback: Callable[..., R] = NoopCallback(),
        **kwargs: Optional[ValueT],
    ) -> R:
        """
        Sends a command to one or many nodes in the cluster
        """
        nodes = self.determine_node(command, **kwargs)
        if nodes and len(nodes) > 1:
            tasks: Dict[str, Coroutine[Any, Any, R]] = {}
            node_arg_mapping = self._split_args_over_nodes(nodes, command, *args)
            node_name_map = {n.name: n for n in nodes}
            for node_name in node_arg_mapping:
                for portion, pargs in enumerate(node_arg_mapping[node_name]):
                    tasks[f"{node_name}:{portion}"] = (
                        self._execute_command_on_single_node(
                            command,
                            *pargs,
                            callback=callback,
                            node=node_name_map[node_name],
                            slots=None,
                            **kwargs,
                        )
                    )

            results = await asyncio.gather(*tasks.values(), return_exceptions=True)
            if self.noreply:
                return None  # type: ignore
            return cast(
                R,
                self._merge_result(command, dict(zip(tasks.keys(), results)), **kwargs),
            )
        else:
            node = None
            slots = None
            if not nodes:
                slots = list(self._determine_slots(command, *args, **kwargs))
            else:
                node = nodes.pop()
            return await self._execute_command_on_single_node(
                command, *args, callback=callback, node=node, slots=slots, **kwargs
            )

    def _split_args_over_nodes(
        self,
        nodes: List[ManagedNode],
        command: bytes,
        *args: ValueT,
    ) -> Dict[str, List[Tuple[ValueT, ...]]]:
        if command in self.split_flags and self.non_atomic_cross_slot:
            keys = KeySpec.extract_keys(command, *args)
            node_arg_mapping: Dict[str, List[Tuple[ValueT, ...]]] = {}
            if keys:
                key_start: int = args.index(keys[0])
                key_end: int = args.index(keys[-1])
                assert (
                    args[key_start : 1 + key_end] == keys
                ), f"Unable to map {command.decode('latin-1')} by keys {keys}"

                for (
                    node_name,
                    key_groups,
                ) in self.connection_pool.nodes.keys_to_nodes_by_slot(*keys).items():
                    for _, node_keys in key_groups.items():
                        node_arg_mapping.setdefault(node_name, []).append(
                            (
                                *args[:key_start],
                                *node_keys,  # type: ignore
                                *args[1 + key_end :],
                            )
                        )
            if self.cache and command not in READONLY_COMMANDS:
                self.cache.invalidate(*keys)
            return node_arg_mapping
        else:
            # This command is not meant to be split across nodes and each node
            # should be called with the same arguments
            return {node.name: [args] for node in nodes}

    async def _execute_command_on_single_node(
        self,
        command: bytes,
        *args: ValueT,
        callback: Callable[..., R] = NoopCallback(),
        node: Optional[ManagedNode] = None,
        slots: Optional[List[int]] = None,
        **kwargs: Optional[ValueT],
    ) -> R:
        redirect_addr = None

        asking = False

        if not node and not slots:
            try_random_node = True
            try_random_type = NodeFlag.PRIMARIES
        else:
            try_random_node = False
            try_random_type = NodeFlag.ALL
        remaining_attempts = int(self.MAX_RETRIES)

        while remaining_attempts > 0:
            remaining_attempts -= 1
            if self.refresh_table_asap and not slots:
                await self
            if asking and redirect_addr:
                node = self.connection_pool.nodes.nodes[redirect_addr]
                r = await self.connection_pool.get_connection_by_node(node)
            elif try_random_node:
                r = await self.connection_pool.get_random_connection(
                    primary=try_random_type == NodeFlag.PRIMARIES
                )
                if slots:
                    try_random_node = False
            elif node:
                r = await self.connection_pool.get_connection_by_node(node)
            elif slots:
                if self.refresh_table_asap:
                    # MOVED
                    node = self.connection_pool.get_primary_node_by_slots(slots)
                else:
                    node = self.connection_pool.get_node_by_slots(slots)
                r = await self.connection_pool.get_connection_by_node(node)
            else:
                continue
            quick_release = self.should_quick_release(command)
            released = False
            try:
                if asking:
                    request = await r.create_request(
                        CommandName.ASKING, noreply=self.noreply, decode=False
                    )
                    await request
                    asking = False

                if (
                    isinstance(self.cache, AbstractCache)
                    and isinstance(self.cache, SupportsClientTracking)
                    and r.tracking_client_id != self.cache.get_client_id(r)
                ):
                    self.cache.reset()
                    await r.update_tracking_client(True, self.cache.get_client_id(r))
                if self.cache and command not in READONLY_COMMANDS:
                    self.cache.invalidate(*KeySpec.extract_keys(command, *args))
                request = await r.create_request(
                    command,
                    *args,
                    noreply=self.noreply,
                    decode=kwargs.get("decode", self._decodecontext.get()),
                    encoding=self._encodingcontext.get(),
                )
                if quick_release and not (self.requires_wait or self.requires_waitaof):
                    released = True
                    self.connection_pool.release(r)

                reply = await request
                response = None
                maybe_wait = [
                    await self._ensure_wait(command, r),
                    await self._ensure_persistence(command, r),
                ]
                if not self.noreply:
                    if isinstance(callback, AsyncPreProcessingCallback):
                        await callback.pre_process(
                            self, reply, version=self.protocol_version, **kwargs
                        )
                    response = callback(
                        reply,
                        version=self.protocol_version,
                        **kwargs,
                    )
                await asyncio.gather(*maybe_wait)
                return response  # type: ignore
            except (RedisClusterException, BusyLoadingError, asyncio.CancelledError):
                raise
            except MovedError as e:
                # Reinitialize on ever x number of MovedError.
                # This counter will increase faster when the same client object
                # is shared between multiple threads. To reduce the frequency you
                # can set the variable 'reinitialize_steps' in the constructor.
                self.refresh_table_asap = True
                await self.connection_pool.nodes.increment_reinitialize_counter()

                node = self.connection_pool.nodes.set_node(
                    e.host, e.port, server_type="primary"
                )
                try_random_node = False
                self.connection_pool.nodes.slots[e.slot_id][0] = node
            except TryAgainError:
                if remaining_attempts < self.MAX_RETRIES / 2:
                    await asyncio.sleep(0.05)
            except AskError as e:
                redirect_addr, asking = f"{e.host}:{e.port}", True
            finally:
                self._ensure_server_version(r.server_version)
                if not released:
                    self.connection_pool.release(r)

        raise ClusterError("Maximum retries exhausted.")

    @overload
    def decoding(
        self, mode: Literal[False], encoding: Optional[str] = None
    ) -> ContextManager[RedisCluster[bytes]]: ...

    @overload
    def decoding(
        self, mode: Literal[True], encoding: Optional[str] = None
    ) -> ContextManager[RedisCluster[str]]: ...

    @contextlib.contextmanager
    @versionadded(version="4.8.0")
    def decoding(
        self, mode: bool, encoding: Optional[str] = None
    ) -> Iterator[RedisCluster[Any]]:
        """
        Context manager to temporarily change the decoding behavior
        of the client

        :param mode: Whether to decode or not
        :param encoding: Optional encoding to use if decoding. If not provided
         the :paramref:`~coredis.RedisCluster.encoding` parameter provided to the client will
         be used.

        Example::

            client = coredis.RedisCluster(decode_responses=True)
            await client.set("fubar", "baz")
            assert await client.get("fubar") == "baz"
            with client.decoding(False):
                assert await client.get("fubar") == b"baz"
                with client.decoding(True):
                    assert await client.get("fubar") == "baz"

        """
        prev_decode = self._decodecontext.get()
        prev_encoding = self._encodingcontext.get()
        self._decodecontext.set(mode)
        self._encodingcontext.set(encoding)
        try:
            yield self
        finally:
            self._decodecontext.set(prev_decode)
            self._encodingcontext.set(prev_encoding)

    def pubsub(
        self,
        ignore_subscribe_messages: bool = False,
        retry_policy: Optional[RetryPolicy] = None,
        **kwargs: Any,
    ) -> ClusterPubSub[AnyStr]:
        """
        Return a Pub/Sub instance that can be used to subscribe to channels or
        patterns in a redis cluster and receive messages that get published to them.

        :param ignore_subscribe_messages: Whether to skip subscription
         acknowledgement messages
        :param retry_policy: An explicit retry policy to use in the subscriber.
        """
        return ClusterPubSub[AnyStr](
            self.connection_pool,
            ignore_subscribe_messages=ignore_subscribe_messages,
            retry_policy=retry_policy,
            **kwargs,
        )

    @versionadded(version="3.6.0")
    def sharded_pubsub(
        self,
        ignore_subscribe_messages: bool = False,
        read_from_replicas: bool = False,
        retry_policy: Optional[RetryPolicy] = None,
        **kwargs: Any,
    ) -> ShardedPubSub[AnyStr]:
        """
        Return a Pub/Sub instance that can be used to subscribe to channels
        in a redis cluster and receive messages that get published to them. The
        implementation returned differs from that returned by :meth:`pubsub`
        as it uses the Sharded Pub/Sub implementation which routes messages
        to cluster nodes using the same algorithm used to assign keys to slots.
        This effectively restricts the propagation of messages to be within the
        shard of a cluster hence affording horizontally scaling the use of Pub/Sub
        with the cluster itself.

        :param ignore_subscribe_messages: Whether to skip subscription
         acknowledgement messages
        :param read_from_replicas: Whether to read messages from replica nodes
        :param retry_policy: An explicit retry policy to use in the subscriber.

        New in :redis-version:`7.0.0`
        """

        return ShardedPubSub[AnyStr](
            self.connection_pool,
            ignore_subscribe_messages=ignore_subscribe_messages,
            read_from_replicas=read_from_replicas,
            retry_policy=retry_policy,
            **kwargs,
        )

    async def pipeline(
        self,
        transaction: Optional[bool] = None,
        watches: Optional[Parameters[StringT]] = None,
        timeout: Optional[float] = None,
    ) -> "coredis.pipeline.ClusterPipeline[AnyStr]":
        """
        Returns a new pipeline object that can queue multiple commands for
        batch execution. Pipelines in cluster mode only provide a subset of the
        functionality of pipelines in standalone mode.

        Specifically:

        - Each command in the pipeline should only access keys on the same node
        - Transactions are disabled by default and are only supported if all
          watched keys route to the same node as where the commands in the multi/exec
          part of the pipeline.

        :param transaction: indicates whether all commands should be executed atomically.
        :param watches: If :paramref:`transaction` is True these keys are watched for external
         changes during the transaction.
        :param timeout: If specified this value will take precedence over
         :paramref:`RedisCluster.stream_timeout`

        """
        await self.connection_pool.initialize()

        from coredis.pipeline import ClusterPipeline

        return ClusterPipeline[AnyStr].proxy(
            client=self,
            transaction=transaction,
            watches=watches,
            timeout=timeout,
        )

    async def transaction(
        self,
        func: Callable[
            ["coredis.pipeline.ClusterPipeline[AnyStr]"],
            Coroutine[Any, Any, Any],
        ],
        *watches: StringT,
        value_from_callable: bool = False,
        watch_delay: Optional[float] = None,
        **kwargs: Any,
    ) -> Any:
        """
        Convenience method for executing the callable :paramref:`func` as a
        transaction while watching all keys specified in :paramref:`watches`.

        :param func: callable should expect a single argument which is a
         :class:`coredis.pipeline.ClusterPipeline` object retrieved by calling
         :meth:`~coredis.RedisCluster.pipeline`.
        :param watches: The keys to watch during the transaction. The keys should route
         to the same node as the keys touched by the commands in :paramref:`func`
        :param value_from_callable: Whether to return the result of transaction or the value
         returned from :paramref:`func`

        .. warning:: Cluster transactions can only be run with commands that
           route to the same slot.

        .. versionchanged:: 4.9.0

           When the transaction is started with :paramref:`watches` the
           :class:`~coredis.pipeline.ClusterPipeline` instance passed to :paramref:`func`
           will not start queuing commands until a call to
           :meth:`~coredis.pipeline.ClusterPipeline.multi` is made. This makes the cluster
           implementation consistent with :meth:`coredis.Redis.transaction`
        """
        async with await self.pipeline(True) as pipe:
            while True:
                try:
                    if watches:
                        await pipe.watch(*watches)
                    func_value = await func(pipe)
                    exec_value = await pipe.execute()
                    return func_value if value_from_callable else exec_value
                except WatchError:
                    if watch_delay is not None and watch_delay > 0:
                        await asyncio.sleep(watch_delay)
                    continue

    async def scan_iter(
        self,
        match: Optional[StringT] = None,
        count: Optional[int] = None,
        type_: Optional[StringT] = None,
    ) -> AsyncIterator[AnyStr]:
        for node in self.primaries:
            cursor = None
            while cursor != 0:
                cursor, data = await node.scan(cursor or 0, match, count, type_)
                for item in data:
                    yield item



from __future__ import annotations

from coredis._utils import EncodingInsensitiveDict
from coredis.response._callbacks import ResponseCallback
from coredis.response._utils import flat_pairs_to_dict
from coredis.typing import (
    AnyStr,
    Dict,
    List,
    Optional,
    ResponseType,
    Set,
    Union,
    ValueT,
)


class ClientTrackingInfoCallback(
    ResponseCallback[
        ResponseType,
        ResponseType,
        Dict[AnyStr, Union[AnyStr, Set[AnyStr], List[AnyStr]]],
    ]
):
    def transform(
        self, response: ResponseType, **options: Optional[ValueT]
    ) -> Dict[AnyStr, Union[AnyStr, Set[AnyStr], List[AnyStr]]]:
        response = EncodingInsensitiveDict(flat_pairs_to_dict(response))
        response["flags"] = set(response["flags"])
        return dict(response)

    def transform_3(
        self, response: ResponseType, **options: Optional[ValueT]
    ) -> Dict[AnyStr, Union[AnyStr, Set[AnyStr], List[AnyStr]]]:
        return response



from __future__ import annotations

import asyncio
import time
import weakref
from typing import TYPE_CHECKING, Any

from coredis.connection import BaseConnection, Connection
from coredis.exceptions import ConnectionError
from coredis.typing import Optional, ResponseType, Set, Tuple, TypeVar

if TYPE_CHECKING:
    import coredis.client

SidecarT = TypeVar("SidecarT", bound="Sidecar")


class Sidecar:
    """
    A sidecar to a redis client that reserves a single connection
    and moves any responses from the socket to a FIFO queue
    """

    def __init__(
        self, push_message_types: Set[bytes], health_check_interval_seconds: int = 5
    ) -> None:
        self._client: Optional[weakref.ReferenceType["coredis.client.Client[Any]"]] = (
            None
        )
        self.messages: asyncio.Queue[ResponseType] = asyncio.Queue()
        self.connection: Optional[Connection] = None
        self.client_id: Optional[int] = None
        self.read_task: Optional[asyncio.Task[None]] = None
        self.push_message_types = push_message_types
        self.health_check_interval = health_check_interval_seconds
        self.health_check_task: Optional[asyncio.Task[None]] = None
        self.last_checkin: float = 0

    @property
    def client(self) -> "Optional[coredis.client.Client[Any]]":
        if self._client:
            return self._client()
        return None  # noqa

    async def start(self: SidecarT, client: "coredis.client.Client[Any]") -> SidecarT:
        self._client = weakref.ref(client, lambda *_: self.stop())
        if not self.connection and self.client:
            self.connection = await self.client.connection_pool.get_connection()
            self.connection.register_connect_callback(self.on_reconnect)
            await self.connection.connect()
            if self.connection.tracking_client_id:  # noqa
                await self.connection.update_tracking_client(False)
        if not self.read_task or self.read_task.done():
            self.read_task = asyncio.create_task(self.__read_loop())
        if not self.health_check_task or self.health_check_task.done():
            self.health_check_task = asyncio.create_task(self.__health_check())
        return self

    def process_message(self, message: ResponseType) -> Tuple[ResponseType, ...]:
        return (message,)  # noqa

    def stop(self) -> None:
        try:
            asyncio.get_running_loop()
            if self.read_task and not self.read_task.done():
                self.read_task.cancel()
            if self.health_check_task and not self.health_check_task.done():
                self.health_check_task.cancel()
        except RuntimeError:
            pass
        if self.connection:
            self.connection.disconnect()
            if self.client and self.connection:  # noqa
                self.client.connection_pool.release(self.connection)
            self.connection = None
            self.client_id = None

    def __del__(self) -> None:
        self.stop()

    async def on_reconnect(self, connection: BaseConnection) -> None:
        self.client_id = connection.client_id
        self.last_checkin = time.monotonic()

    async def __health_check(self) -> None:
        while True:
            try:
                if self.connection:
                    await self.connection.send_command(b"PING")
                await asyncio.sleep(self.health_check_interval)
            except asyncio.CancelledError:
                break

    async def __read_loop(self) -> None:
        while self.connection:
            try:
                response = await self.connection.fetch_push_message(
                    decode=False, push_message_types=self.push_message_types
                )
                self.last_checkin = time.monotonic()
                if response == b"PONG" or b"pong" in response:  # type: ignore
                    continue
                for m in self.process_message(response):
                    self.messages.put_nowait(m)
            except asyncio.CancelledError:
                break
            except ConnectionError:
                if self.client and self.connection:
                    self.client.connection_pool.release(self.connection)
                self.connection = None

                if self.client:
                    await self.start(self.client)



from __future__ import annotations

from typing import Any

from deprecated.sphinx import versionadded

from coredis._utils import EncodingInsensitiveDict, nativestr
from coredis.client import Client
from coredis.exceptions import (
    ResponseError,
    StreamConsumerInitializationError,
    StreamDuplicateConsumerGroupError,
)
from coredis.response.types import StreamEntry
from coredis.tokens import PureToken
from coredis.typing import (
    AnyStr,
    ClassVar,
    Dict,
    Generator,
    Generic,
    KeyT,
    List,
    Optional,
    Parameters,
    Set,
    StringT,
    Tuple,
    TypedDict,
    ValueT,
)


class StreamParameters(TypedDict):
    #: Starting ``identifier`` for the consumer. If not present it will start
    #: from the latest entry
    identifier: StringT


class State(TypedDict, total=False):
    identifier: Optional[StringT]
    pending: Optional[bool]


class Consumer(Generic[AnyStr]):
    state: Dict[KeyT, State]
    DEFAULT_START_ID: ClassVar[bytes] = b"0-0"

    def __init__(
        self,
        client: Client[AnyStr],
        streams: Parameters[KeyT],
        buffer_size: int = 0,
        timeout: Optional[int] = 0,
        **stream_parameters: StreamParameters,
    ):
        """
        Standalone stream consumer that starts reading from the latest entry
        of each stream provided in :paramref:`streams`.

        The latest entry is determined by calling :meth:`coredis.Redis.xinfo_stream`
        and using the :data:`last-entry` attribute
        at the point of initializing the consumer instance or on first fetch (whichever comes
        first). If the stream(s) do not exist at the time of consumer creation, the
        consumer will simply start from the minimum identifier (``0-0``)

        :param client: The redis client to use
        :param streams: the stream identifiers to consume from
        :param buffer_size: Size of buffer (per stream) to maintain. This
         translates to the maximum number of stream entries that are fetched
         on each request to redis.
        :param timeout: Maximum amount of time in milliseconds to block for new
         entries to appear on the streams the consumer is reading from.
        :param stream_parameters: Mapping of optional parameters to use
         by stream for the streams provided in :paramref:`streams`.
        """
        self.client: Client[AnyStr] = client
        self.streams: Set[KeyT] = set(streams)
        self.state: Dict[StringT, State] = EncodingInsensitiveDict(
            {stream: stream_parameters.get(nativestr(stream), {}) for stream in streams}
        )
        self.buffer: Dict[AnyStr, List[StreamEntry]] = EncodingInsensitiveDict({})
        self.buffer_size = buffer_size
        self.timeout = timeout
        self._initialized = False
        self._initialized_streams: Dict[StringT, bool] = {}

    def chunk_streams(self) -> List[Dict[ValueT, StringT]]:
        import coredis.client

        if isinstance(self.client, coredis.client.RedisCluster):
            return [
                {
                    stream: self.state[stream].get("identifier", None)
                    or self.DEFAULT_START_ID
                }
                for stream in self.streams
            ]
        else:
            return [
                {
                    stream: self.state[stream].get("identifier", None)
                    or self.DEFAULT_START_ID
                    for stream in self.streams
                }
            ]

    async def initialize(self, partial: bool = False) -> "Consumer[AnyStr]":
        if self._initialized and not partial:
            return self

        for stream in self.streams:
            if partial and self._initialized_streams.get(stream):
                continue
            try:
                info = await self.client.xinfo_stream(stream)
                if info:
                    last_entry = info["last-entry"]
                    if last_entry:
                        self.state[stream].setdefault(
                            "identifier", last_entry.identifier
                        )
            except ResponseError:
                pass
            self._initialized_streams[stream] = True
        self._initialized = True
        return self

    async def add_stream(
        self, stream: StringT, identifier: Optional[StringT] = None
    ) -> bool:
        """
        Adds a new stream identifier to this consumer

        :param stream: The stream identifier
        :return: ``True`` if the stream was added successfully, ``False`` otherwise
        """
        self.streams.add(stream)
        self.state.setdefault(stream, {"identifier": identifier} if identifier else {})
        await self.initialize(partial=True)
        return stream in self._initialized_streams

    def __await__(self) -> Generator[Any, None, Consumer[AnyStr]]:
        return self.initialize().__await__()

    def __aiter__(self) -> Consumer[AnyStr]:
        """
        Returns the instance of the consumer itself which can be iterated over
        """
        return self

    async def __anext__(self) -> Tuple[AnyStr, StreamEntry]:
        """
        Returns the next available stream entry available from any of
        :paramref:`Consumer.streams`.

        :raises: :exc:`StopIteration` if no more entries are available
        """
        stream, entry = await self.get_entry()
        if not (stream and entry):
            raise StopAsyncIteration()
        return stream, entry

    async def get_entry(self) -> Tuple[Optional[AnyStr], Optional[StreamEntry]]:
        """
        Fetches the next available entry from the streams specified in
        :paramref:`Consumer.streams`. If there were any entries
        previously fetched and buffered, they will be returned before
        making a new request to the server.
        """
        await self.initialize()
        cur = None
        cur_stream = None
        for stream, buffer_entries in list(self.buffer.items()):
            if buffer_entries:
                cur_stream, cur = stream, self.buffer[stream].pop(0)
                break
        else:
            consumed_entries: Dict[AnyStr, Tuple[StreamEntry, ...]] = {}
            for chunk in self.chunk_streams():
                consumed_entries.update(
                    await self.client.xread(
                        chunk,
                        count=self.buffer_size + 1,
                        block=(
                            self.timeout
                            if (self.timeout and self.timeout > 0)
                            else None
                        ),
                    )
                    or {}
                )
            for stream, entries in consumed_entries.items():
                if entries:
                    if not cur:
                        cur = entries[0]
                        cur_stream = stream
                        if entries[1:]:
                            self.buffer.setdefault(stream, []).extend(entries[1:])
                    else:
                        self.buffer.setdefault(stream, []).extend(entries)
        if cur and cur_stream:
            self.state[cur_stream]["identifier"] = cur.identifier
        return cur_stream, cur


class GroupConsumer(Consumer[AnyStr]):
    DEFAULT_START_ID: ClassVar[bytes] = b">"

    def __init__(
        self,
        client: Client[AnyStr],
        streams: Parameters[KeyT],
        group: StringT,
        consumer: StringT,
        buffer_size: int = 0,
        auto_create: bool = True,
        auto_acknowledge: bool = False,
        start_from_backlog: bool = False,
        timeout: Optional[int] = None,
        **stream_parameters: StreamParameters,
    ):
        """
        A member of a stream consumer group. The consumer has an identical
        interface as :class:`coredis.stream.Consumer`.

        :param client: The redis client to use
        :param streams: The stream identifiers to consume from
        :param group: The name of the group this consumer is part of
        :param consumer: The unique name (within :paramref:`group`) of the consumer
        :param auto_create: If True the group will be created upon initialization
         or first fetch if it doesn't already exist.
        :param auto_acknowledge: If ``True`` the stream entries fetched will be fetched
         without needing to be acknowledged with :meth:`coredis.Redis.xack` to remove
         them from the pending entries list.
        :param start_from_backlog: If ``True`` the consumer will start by fetching any pending
         entries from the pending entry list before considering any new messages
         not seen by any other consumer in the :paramref:`group`
        :param buffer_size: Size of buffer (per stream) to maintain. This
         translates to the maximum number of stream entries that are fetched
         on each request to redis.
        :param timeout: Maximum amount of time to block for new
         entries to appear on the streams the consumer is reading from.
        :param stream_parameters: Mapping of optional parameters to use
         by stream for the streams provided in :paramref:`streams`.


        .. warning:: Providing an ``identifier`` in ``stream_parameters`` has a different
           meaning for a group consumer. If the value is any valid identifier other than ``>``
           the consumer will only access the history of pending messages. That is, the set of
           messages that were delivered to this consumer (identified by :paramref:`consumer`)
           and never acknowledged.
        """
        super().__init__(
            client,  # type: ignore[arg-type]
            streams,
            buffer_size,
            timeout,
            **stream_parameters,
        )
        self.group = group
        self.consumer = consumer
        self.auto_create = auto_create
        self.auto_acknowledge = auto_acknowledge
        self.start_from_backlog = start_from_backlog

    async def initialize(self, partial: bool = False) -> "GroupConsumer[AnyStr]":
        if not self._initialized or partial:
            group_presence: Dict[KeyT, bool] = {
                stream: stream in self._initialized_streams for stream in self.streams
            }
            for stream in self.streams:
                try:
                    if self._initialized_streams.get(stream):
                        continue
                    group_presence[stream] = (
                        len(
                            [
                                info
                                for info in [
                                    EncodingInsensitiveDict(d)
                                    for d in await self.client.xinfo_groups(stream)
                                ]
                                if nativestr(info["name"]) == self.group
                            ]
                        )
                        == 1
                    )
                    if group_presence[stream] and self.start_from_backlog:
                        self.state[stream]["pending"] = True
                        self.state[stream]["identifier"] = "0-0"
                except ResponseError:
                    self.state[stream].setdefault("identifier", ">")

            if not (self.auto_create or all(group_presence.values())):
                missing_streams = self.streams - {
                    k for k in group_presence if not group_presence[k]
                }
                raise StreamConsumerInitializationError(
                    f"Consumer group: {self.group!r} does not exist for streams: {missing_streams}"
                )
            for stream in self.streams:
                if self.auto_create and not group_presence.get(stream):
                    try:
                        await self.client.xgroup_create(
                            stream, self.group, PureToken.NEW_ID, mkstream=True
                        )
                    except StreamDuplicateConsumerGroupError:  # noqa
                        pass
                self._initialized_streams[stream] = True
                self.state[stream].setdefault("identifier", ">")

            self._initialized = True
        return self

    @versionadded(version="4.12.0")
    async def add_stream(
        self, stream: StringT, identifier: Optional[StringT] = ">"
    ) -> bool:
        """
        Adds a new stream identifier to this consumer

        :param stream: The stream identifier
        :param identifier: The identifier to start consuming from. For group
         consumers this should almost always be ``>`` (the default).

        :return: ``True`` if the stream was added successfully, ``False`` otherwise
        """
        return await super().add_stream(stream, identifier)

    def __await__(self) -> Generator[Any, None, GroupConsumer[AnyStr]]:
        return self.initialize().__await__()

    def __aiter__(self) -> GroupConsumer[AnyStr]:
        """
        Returns the instance of the consumer itself which can be iterated over
        """
        return self

    async def get_entry(self) -> Tuple[Optional[AnyStr], Optional[StreamEntry]]:
        """
        Fetches the next available entry from the streams specified in
        :paramref:`GroupConsumer.streams`. If there were any entries
        previously fetched and buffered, they will be returned before
        making a new request to the server.
        """
        await self.initialize()

        cur = None
        cur_stream = None
        for stream, buffer_entries in list(self.buffer.items()):
            if buffer_entries:
                cur_stream, cur = stream, self.buffer[stream].pop(0)
                break
        else:
            consumed_entries: Dict[AnyStr, Tuple[StreamEntry, ...]] = {}
            for chunk in self.chunk_streams():
                consumed_entries.update(
                    await self.client.xreadgroup(
                        self.group,
                        self.consumer,
                        count=self.buffer_size + 1,
                        block=(
                            self.timeout
                            if (self.timeout and self.timeout > 0)
                            else None
                        ),
                        noack=self.auto_acknowledge,
                        streams=chunk,
                    )
                    or {}
                )
            for stream, entries in consumed_entries.items():
                if entries:
                    if not cur:
                        cur = entries[0]
                        cur_stream = stream
                        if entries[1:]:
                            self.buffer.setdefault(stream, []).extend(entries[1:])

                    else:
                        self.buffer.setdefault(stream, []).extend(entries)
                    if self.state[stream].get("pending"):
                        self.state[stream]["identifier"] = entries[-1].identifier
                else:
                    if self.state[stream].get("pending"):
                        self.state[stream].pop("identifier", None)
                        self.state[stream].pop("pending", None)
                        if not cur:
                            return await self.get_entry()
        return cur_stream, cur



from __future__ import annotations

from typing import SupportsFloat, cast

from coredis.response._callbacks import ResponseCallback
from coredis.response.types import ScoredMember, ScoredMembers
from coredis.typing import (
    AnyStr,
    Generic,
    List,
    Optional,
    ResponsePrimitive,
    ResponseType,
    Tuple,
    Union,
    ValueT,
)


class ZRankCallback(
    ResponseCallback[
        Optional[Union[int, List[ResponsePrimitive]]],
        Optional[Union[int, List[ResponsePrimitive]]],
        Optional[Union[int, Tuple[int, float]]],
    ],
):
    def transform(
        self,
        response: Optional[Union[int, List[ResponsePrimitive]]],
        **options: Optional[ValueT],
    ) -> Optional[Union[int, Tuple[int, float]]]:
        if options.get("withscore"):
            return (response[0], float(response[1])) if response else None
        else:
            return cast(Optional[int], response)

    def transform_3(
        self,
        response: Optional[Union[int, List[ResponsePrimitive]]],
        **options: Optional[ValueT],
    ) -> Optional[Union[int, Tuple[int, float]]]:
        if options.get("withscore"):
            return (response[0], response[1]) if response else None
        else:
            return cast(Optional[int], response)


class ZMembersOrScoredMembers(
    ResponseCallback[
        List[Union[AnyStr, List[ResponsePrimitive]]],
        List[Union[AnyStr, List[ResponsePrimitive]]],
        Tuple[Union[AnyStr, ScoredMember], ...],
    ],
):
    def transform(
        self,
        response: List[Union[AnyStr, List[ResponsePrimitive]]],
        **options: Optional[ValueT],
    ) -> Tuple[Union[AnyStr, ScoredMember], ...]:
        if not response:
            return ()
        elif options.get("withscores"):
            it = iter(cast(List[AnyStr], response))
            return tuple(ScoredMember(*v) for v in zip(it, map(float, it)))
        else:
            return cast(Tuple[AnyStr, ...], tuple(response))

    def transform_3(
        self,
        response: List[Union[AnyStr, List[ResponsePrimitive]]],
        **options: Optional[ValueT],
    ) -> Tuple[Union[AnyStr, ScoredMember], ...]:
        if options.get("withscores"):
            return tuple(
                ScoredMember(*v) for v in cast(List[Tuple[AnyStr, float]], response)
            )
        else:
            return cast(Tuple[AnyStr, ...], tuple(response))


class ZSetScorePairCallback(
    ResponseCallback[
        Optional[List[ResponsePrimitive]],
        Optional[List[Union[ResponsePrimitive, List[ResponsePrimitive]]]],
        Optional[Union[ScoredMember, ScoredMembers]],
    ],
    Generic[AnyStr],
):
    def transform(
        self, response: Optional[List[ResponsePrimitive]], **options: Optional[ValueT]
    ) -> Optional[Union[ScoredMember, ScoredMembers]]:
        if not response:
            return None

        if not (options.get("withscores") or options.get("count")):
            return ScoredMember(
                cast(AnyStr, response[0]), float(cast(SupportsFloat, response[1]))
            )

        it = iter(response)
        return tuple(ScoredMember(*v) for v in zip(it, map(float, it)))

    def transform_3(
        self,
        response: Optional[List[Union[ResponsePrimitive, List[ResponsePrimitive]]]],
        **options: Optional[ValueT],
    ) -> Optional[Union[ScoredMember, ScoredMembers]]:
        if not response:
            return None

        if not (options.get("withscores") or options.get("count")):
            return ScoredMember(*cast(Tuple[AnyStr, float], response))

        return tuple(
            ScoredMember(*v) for v in cast(List[Tuple[AnyStr, float]], response)
        )


class ZMPopCallback(
    ResponseCallback[
        Optional[List[ResponseType]],
        Optional[List[ResponseType]],
        Optional[Tuple[AnyStr, ScoredMembers]],
    ],
    Generic[AnyStr],
):
    def transform(
        self, response: Optional[List[ResponseType]], **options: Optional[ValueT]
    ) -> Optional[Tuple[AnyStr, ScoredMembers]]:
        r = cast(Tuple[AnyStr, List[Tuple[AnyStr, int]]], response)
        if r:
            return r[0], tuple(ScoredMember(v[0], float(v[1])) for v in r[1])

        return None


class ZMScoreCallback(
    ResponseCallback[
        List[ResponsePrimitive], List[ResponsePrimitive], Tuple[Optional[float], ...]
    ]
):
    def transform(
        self, response: List[ResponsePrimitive], **options: Optional[ValueT]
    ) -> Tuple[Optional[float], ...]:
        return tuple(score if score is None else float(score) for score in response)


class ZScanCallback(
    ResponseCallback[List[ResponseType], List[ResponseType], Tuple[int, ScoredMembers]],
    Generic[AnyStr],
):
    def transform(
        self, response: List[ResponseType], **options: Optional[ValueT]
    ) -> Tuple[int, ScoredMembers]:
        cursor, r = cast(Tuple[int, List[AnyStr]], response)
        it = iter(r)
        return int(cursor), tuple(
            ScoredMember(*cast(Tuple[AnyStr, float], v))
            for v in zip(it, map(float, it))
        )


class ZRandMemberCallback(
    ResponseCallback[
        Optional[Union[AnyStr, List[ResponsePrimitive]]],
        Optional[Union[AnyStr, List[ResponsePrimitive]]],
        Optional[Union[AnyStr, ScoredMembers]],
    ]
):
    def transform(
        self,
        response: Optional[Union[AnyStr, List[ResponsePrimitive]]],
        **options: Optional[ValueT],
    ) -> Optional[Union[AnyStr, ScoredMembers]]:
        if not (response and options.get("withscores")):
            return response

        it = iter(response)
        return tuple(ScoredMember(*v) for v in zip(it, map(float, it)))

    def transform_3(
        self,
        response: Optional[Union[AnyStr, List[ResponsePrimitive]]],
        **options: Optional[ValueT],
    ) -> Optional[Union[AnyStr, ScoredMembers]]:
        if not (response and options.get("withscores")):
            return response

        return tuple(ScoredMember(*v) for v in response)


class BZPopCallback(
    ResponseCallback[
        Optional[List[ResponsePrimitive]],
        Optional[List[ResponsePrimitive]],
        Optional[Tuple[AnyStr, AnyStr, float]],
    ]
):
    def transform(
        self, response: Optional[List[ResponsePrimitive]], **options: Optional[ValueT]
    ) -> Optional[Tuple[AnyStr, AnyStr, float]]:
        if response:
            return response[0], response[1], float(response[2])
        return None


class ZAddCallback(
    ResponseCallback[ResponsePrimitive, Union[int, float], Union[int, float]]
):
    def transform(
        self, response: ResponsePrimitive, **options: Optional[ValueT]
    ) -> Union[int, float]:
        if options.get("condition"):
            return float(response)
        return int(response)

    def transform_3(
        self, response: Union[int, float], **options: Optional[ValueT]
    ) -> Union[int, float]:
        return response



from __future__ import annotations

import functools
import textwrap
from abc import ABCMeta
from typing import TYPE_CHECKING, Any, cast

from packaging import version

from coredis.config import Config

from .._protocols import AbstractExecutor
from ..commands._utils import redis_command_link
from ..commands._wrappers import (
    CacheConfig,
    ClusterCommandConfig,
    CommandCache,
    CommandDetails,
)
from ..commands.constants import CommandFlag, CommandGroup, CommandName
from ..exceptions import CommandSyntaxError, ModuleCommandNotSupportedError
from ..globals import COMMAND_FLAGS, MODULE_GROUPS, MODULES, READONLY_COMMANDS
from ..response._callbacks import NoopCallback
from ..typing import (
    AnyStr,
    Callable,
    ClassVar,
    Coroutine,
    Dict,
    Generic,
    Optional,
    P,
    R,
    Set,
    Tuple,
    Type,
    ValueT,
    add_runtime_checks,
)

if TYPE_CHECKING:
    import coredis.client


async def ensure_compatibility(
    client: "coredis.client.Client[Any]",
    module: str,
    command_details: CommandDetails,
    kwargs: Dict[str, Any],
) -> None:
    if (
        Config.optimized
        or not command_details.version_introduced
        or not getattr(client, "verify_version", False)
        or getattr(client, "noreply", False)
    ):
        return
    if command_details.version_introduced:
        module_version = await client.get_server_module_version(module)
        if module_version and command_details.version_introduced <= module_version:
            if command_details.arguments and set(
                command_details.arguments.keys()
            ).intersection(kwargs.keys()):
                for argument, version_introduced in command_details.arguments.items():
                    if version_introduced and version_introduced > module_version:
                        raise CommandSyntaxError(
                            {argument},
                            (
                                f"{command_details.command.decode('latin-1')} with `{argument}` "
                                f"is not supported in {module} version {module_version}"
                            ),
                        )
            return
        raise ModuleCommandNotSupportedError(
            command_details.command.decode("latin-1"),
            module,
            str(module_version) if module_version else "",
        )


def module_command(
    command_name: CommandName,
    module: "Type[Module[Any]]",
    group: CommandGroup,
    flags: Optional[Set[CommandFlag]] = None,
    cluster: ClusterCommandConfig = ClusterCommandConfig(),
    cache_config: Optional[CacheConfig] = None,
    version_introduced: Optional[str] = None,
    version_deprecated: Optional[str] = None,
    arguments: Optional[Dict[str, Dict[str, str]]] = None,
) -> Callable[
    [Callable[P, Coroutine[Any, Any, R]]], Callable[P, Coroutine[Any, Any, R]]
]:
    command_details = CommandDetails(
        command_name,
        group,
        version.Version(version_introduced) if version_introduced else None,
        version.Version(version_deprecated) if version_deprecated else None,
        arguments,
        cluster or ClusterCommandConfig(),
        cache_config,
        flags or set(),
        None,
    )

    def wrapper(
        func: Callable[P, Coroutine[Any, Any, R]]
    ) -> Callable[P, Coroutine[Any, Any, R]]:
        runtime_checkable = add_runtime_checks(func)
        command_cache = CommandCache(command_name, cache_config)
        if flags and CommandFlag.READONLY in flags:
            READONLY_COMMANDS.add(command_name)
        COMMAND_FLAGS[command_name] = flags or set()

        @functools.wraps(func)
        async def wrapped(*args: P.args, **kwargs: P.kwargs) -> R:
            from coredis.client import Redis, RedisCluster

            mg = cast(ModuleGroup[bytes], args[0])
            client = cast("coredis.client.Client[Any]", mg.client)
            is_regular_client = isinstance(client, (Redis, RedisCluster))
            runtime_checking = (
                not getattr(client, "noreply", None) and is_regular_client
            )
            callable = runtime_checkable if runtime_checking else func
            await ensure_compatibility(client, module.NAME, command_details, kwargs)
            async with command_cache(callable, *args, **kwargs) as response:
                return response

        wrapped.__doc__ = textwrap.dedent(wrapped.__doc__ or "")
        if group:
            wrapped.__doc__ = f"""
{wrapped.__doc__}

{module.FULL_NAME} command documentation: {redis_command_link(command_name)}
            """
        if (
            (version_introduced and version_introduced != "1.0.0")
            or version_deprecated
            or command_details.arguments
        ):
            wrapped.__doc__ += """
Compatibility:
"""

            if version_introduced and version_introduced != "1.0.0":
                wrapped.__doc__ += f"""
- New in {module.FULL_NAME} version: `{version_introduced}`
                """
            if version_deprecated:
                wrapped.__doc__ += f"""
- Deprecated in {module.FULL_NAME} version: `{version_deprecated}`
        """
            if command_details.arguments:
                for argument, min_version in command_details.arguments.items():
                    wrapped.__doc__ += f"""
- :paramref:`{argument}`: New in {module.FULL_NAME} version `{min_version}`
                    """
        if cache_config:
            wrapped.__doc__ += """
.. hint:: Supports client side caching
"""
        setattr(wrapped, "__coredis_command", command_details)
        setattr(wrapped, "__coredis_module", module)
        return wrapped

    return wrapper


class ModuleRegistry(ABCMeta):
    """
    :meta private:
    """

    NAME: str
    FULL_NAME: str
    DESCRIPTION: str
    DOCUMENTATION_URL: str
    COMMAND_GROUPS: Dict[CommandGroup, Type[ModuleGroup[Any]]]

    def __new__(
        cls, name: str, bases: Tuple[type, ...], namespace: Dict[str, object]
    ) -> ModuleRegistry:
        if "COMMAND_GROUPS" not in namespace:
            namespace["COMMAND_GROUPS"] = {}
        kls = super().__new__(cls, name, bases, namespace)
        if getattr(kls, "NAME", None):
            MODULES[kls.NAME] = kls
            kls.__doc__ = textwrap.dedent(kls.__doc__ or "")
            kls.__doc__ += f"""
Type representation of the `{kls.FULL_NAME} <{kls.DOCUMENTATION_URL}>`__ module.
The class isn't meant to be used directly and exists as a convenient way to capture
module attributes for documentation & internal use.
            """
        return kls


class ModuleGroupRegistry(ABCMeta):
    """
    :meta private:
    """

    MODULE: Type[Module[Any]]
    COMMAND_GROUP: CommandGroup

    def __new__(
        cls, name: str, bases: Tuple[type, ...], namespace: Dict[str, object]
    ) -> ModuleGroupRegistry:
        kls = super().__new__(cls, name, bases, namespace)
        if getattr(kls, "MODULE", None):
            MODULE_GROUPS.add(kls)
            kls.MODULE.COMMAND_GROUPS[kls.COMMAND_GROUP] = cast(
                Type[ModuleGroup[Any]], kls
            )
            original_doc = textwrap.dedent(kls.__doc__ or "")
            kls.__doc__ = f"""
Container for the commands in the ``{kls.COMMAND_GROUP.value}`` command group of the
`{kls.MODULE.FULL_NAME} <{kls.MODULE.DOCUMENTATION_URL}>`__ module.

{original_doc}
            """
        return kls


class Module(Generic[AnyStr], metaclass=ModuleRegistry):
    """
    Base class for Redis module implementations.
    This class isn't meant to be used directly by the end
    user, and exists to provide a convenient way to document and group
    redis modules and the command groups that they expose.
    """

    NAME: ClassVar[str]
    """
    The name of the module as reported by ``MODULES LIST``
    """
    FULL_NAME: ClassVar[str]
    """
    The common name used to refer to the module if it differs from
    the internal name
    """
    DESCRIPTION: ClassVar[str]
    """
    A brief description of what the module does
    """
    DOCUMENTATION_URL: ClassVar[str]
    """
    A link to the module documentation
    """

    COMMAND_GROUPS: ClassVar[Dict[CommandGroup, Type[ModuleGroup[Any]]]]
    """
    Mapping of command groups that implement this module. This is auto
    populated by the :class:`ModuleGroupRegistry` metaclass.
    """


class ModuleGroup(Generic[AnyStr], metaclass=ModuleGroupRegistry):
    """
    Base class for Redis module command groups
    """

    MODULE: ClassVar[Type[Module[Any]]]
    """
    The module to which this command group belongs to
    
    :meta private: 
    """
    COMMAND_GROUP: ClassVar[CommandGroup]
    """
    The command group this class implements

    :meta private: 
    """

    def __init__(self, client: AbstractExecutor):
        self.client = client

    async def execute_module_command(
        self,
        command: bytes,
        *args: ValueT,
        callback: Callable[..., R] = NoopCallback(),
        **options: Optional[ValueT],
    ) -> R:
        return cast(
            R,
            await self.client.execute_command(
                command, *args, callback=callback, **options
            ),
        )



from __future__ import annotations

from collections import defaultdict
from typing import TYPE_CHECKING

from coredis.commands.constants import CommandFlag
from coredis.typing import Dict, Set

if TYPE_CHECKING:
    from coredis.modules.base import ModuleGroupRegistry, ModuleRegistry

#: Populated by the @redis_command wrapper
READONLY_COMMANDS: Set[bytes] = set()
#: Populated by the @redis_command wrapper
COMMAND_FLAGS: Dict[bytes, Set[CommandFlag]] = defaultdict(lambda: set())

#: Populated by ModuleGroupRegistry
MODULE_GROUPS: Set[ModuleGroupRegistry] = set()

#: Populated by ModuleRegistry
MODULES: Dict[str, ModuleRegistry] = {}



from __future__ import annotations

from .basic import Client, Redis
from .cluster import RedisCluster
from .keydb import KeyDB, KeyDBCluster

__all__ = ["Client", "Redis", "RedisCluster", "KeyDB", "KeyDBCluster"]



from __future__ import annotations

from datetime import datetime

from coredis.exceptions import DataError, NoKeyError, RedisError
from coredis.response._callbacks import DateTimeCallback, ResponseCallback
from coredis.typing import (
    AnyStr,
    List,
    Optional,
    ResponseType,
    StringT,
    Tuple,
    TypeGuard,
    Union,
    ValueT,
)


class SortCallback(
    ResponseCallback[
        Union[int, List[AnyStr]],
        Union[int, List[AnyStr]],
        Union[int, Tuple[AnyStr, ...]],
    ]
):
    def transform(
        self, response: Union[int, List[AnyStr]], **options: Optional[ValueT]
    ) -> Union[int, Tuple[AnyStr, ...]]:
        if isinstance(response, list):
            return tuple(response)
        return response


class ScanCallback(
    ResponseCallback[
        List[ResponseType], List[ResponseType], Tuple[int, Tuple[AnyStr, ...]]
    ]
):
    def guard(
        self, response: List[ResponseType]
    ) -> TypeGuard[Tuple[StringT, List[AnyStr]]]:
        return isinstance(response[0], (str, bytes)) and isinstance(response[1], list)

    def transform(
        self, response: List[ResponseType], **options: Optional[ValueT]
    ) -> Tuple[int, Tuple[AnyStr, ...]]:
        assert self.guard(response)
        cursor, r = response
        return int(cursor), tuple(r)


class ExpiryCallback(DateTimeCallback):
    def transform(self, response: int, **options: Optional[ValueT]) -> datetime:
        if response > 0:
            return super().transform(response, **options)
        else:
            if response == -2:
                raise NoKeyError()
            elif response == -1:
                raise DataError()
            else:
                raise RedisError()



from __future__ import annotations

import enum
from typing import Any

from wrapt import ObjectProxy

from coredis.typing import (
    Hashable,
    Iterable,
    List,
    Mapping,
    Optional,
    ResponseType,
    Set,
    StringT,
    Tuple,
    TypeVar,
    Union,
)

T = TypeVar("T")
U = TypeVar("U")


class EncodingInsensitiveDict(ObjectProxy):  # type: ignore
    def __init__(
        self,
        dict: Optional[Union[Mapping[Any, Any], ResponseType]] = None,
        encoding: str = "utf-8",
    ):
        d = dict or {}
        super().__init__(d)
        self._self_encoding = encoding

    def __getitem__(self, item: StringT) -> Any:
        if isinstance(item, str) and item not in self.__wrapped__:
            return self.__wrapped__.get(
                item, self.__wrapped__.get(item.encode(self._self_encoding))
            )
        elif isinstance(item, bytes) and item not in self.__wrapped__:
            return self.__wrapped__.get(
                item, self.__wrapped__.get(item.decode(self._self_encoding))
            )
        return self.__wrapped__[item]

    def get(self, item: StringT, default: Optional[object] = None) -> Any:
        return self.__getitem__(item) or default

    def pop(self, item: StringT, default: Optional[object] = None) -> Any:
        if item in self.__wrapped__:
            return self.__wrapped__.pop(item)
        if isinstance(item, str):
            return self.__wrapped__.pop(item.encode(self._self_encoding), default)

    def clear(self) -> None:
        self.__wrapped__.clear()

    def update(self, updates: Mapping[Any, Any]) -> None:
        self.__wrapped__.update(updates)

    def __setitem__(self, item: StringT, value: object) -> None:
        if item in self.__wrapped__:
            self.__wrapped__[item] = value
        elif (
            isinstance(item, str)
            and item.encode(self._self_encoding) in self.__wrapped__
        ):
            self.__wrapped__[item.encode(self._self_encoding)] = value
        elif (
            isinstance(item, bytes)
            and item.decode(self._self_encoding) in self.__wrapped__
        ):
            self.__wrapped__[item.decode(self._self_encoding)] = value
        else:
            self.__wrapped__[item] = value

    def __contains__(self, key: StringT) -> bool:
        if isinstance(key, str):
            return (
                key in self.__wrapped__
                or key.encode(self._self_encoding) in self.__wrapped__
            )
        elif isinstance(key, bytes):
            return (
                key in self.__wrapped__
                or key.decode(self._self_encoding) in self.__wrapped__
            )
        return key in self.__wrapped__

    def __repr__(self) -> str:
        return repr(self.__wrapped__)


@enum.unique
class CaseAndEncodingInsensitiveEnum(bytes, enum.Enum):
    __decoded: Set[StringT]

    @property
    def variants(self) -> Set[StringT]:
        if not hasattr(self, "__decoded"):
            decoded = str(self)
            self.__decoded = {
                self.value.lower(),  # type: ignore
                self.value,  # type: ignore
                decoded.lower(),
                decoded.upper(),
            }
        return self.__decoded

    def __eq__(self, other: object) -> bool:
        """
        Since redis tokens are case insensitive allow mixed case
        Additionally allow strings to be passed in instead of
        bytes.
        """

        if other:
            if isinstance(other, self.__class__):
                return bool(self.value == other.value)
            else:
                return other in self.variants
        return False

    def __str__(self) -> str:
        return self.decode("latin-1")

    def __hash__(self) -> int:
        return hash(self.value)


def b(x: ResponseType, encoding: Optional[str] = None) -> bytes:
    if isinstance(x, bytes):
        return x
    if not isinstance(x, str):
        _v = str(x)
    else:
        _v = x
    return _v.encode(encoding) if encoding else _v.encode()


def defaultvalue(value: Optional[U], default: T) -> Union[U, T]:
    return default if value is None else value


def nativestr(x: ResponseType, encoding: str = "utf-8") -> str:
    if isinstance(x, (str, bytes)):
        return x if isinstance(x, str) else x.decode(encoding, "replace")
    elif isinstance(x, (int, float, bool)):
        return str(x)
    raise ValueError(f"Unable to cast {x} to string")


def tuples_to_flat_list(nested_list: Iterable[Tuple[T, ...]]) -> List[T]:
    return [item for sublist in nested_list for item in sublist]


def dict_to_flat_list(
    mapping: Mapping[T, U], reverse: bool = False
) -> List[Union[T, U]]:
    e1: List[Union[T, U]] = list(mapping.keys())
    e2: List[Union[T, U]] = list(mapping.values())

    ret: List[Union[T, U]] = []

    if reverse:
        e1, e2 = e2, e1

    for idx, e in enumerate(e1):
        ret.append(e)
        ret.append(e2[idx])

    return ret


def make_hashable(*args: Any) -> Tuple[Hashable, ...]:
    return tuple(
        (
            (
                tuple(make_hashable(v)[0] for v in a)
                if isinstance(a, (tuple, list))
                else (
                    frozenset(make_hashable(v)[0] for v in a)
                    if isinstance(a, set)
                    else (
                        tuple((k, make_hashable(v)[0]) for k, v in a.items())
                        if isinstance(a, dict)
                        # this will fail downstream if `a` is not hashable
                        else a
                    )
                )
            )
            for a in args
        )
    )


# ++++++++++ cluster utils ++++++++++++++

try:
    from coredis.speedups import crc16, hash_slot
except ImportError:
    x_mode_m_crc16_lookup = [
        0x0000,
        0x1021,
        0x2042,
        0x3063,
        0x4084,
        0x50A5,
        0x60C6,
        0x70E7,
        0x8108,
        0x9129,
        0xA14A,
        0xB16B,
        0xC18C,
        0xD1AD,
        0xE1CE,
        0xF1EF,
        0x1231,
        0x0210,
        0x3273,
        0x2252,
        0x52B5,
        0x4294,
        0x72F7,
        0x62D6,
        0x9339,
        0x8318,
        0xB37B,
        0xA35A,
        0xD3BD,
        0xC39C,
        0xF3FF,
        0xE3DE,
        0x2462,
        0x3443,
        0x0420,
        0x1401,
        0x64E6,
        0x74C7,
        0x44A4,
        0x5485,
        0xA56A,
        0xB54B,
        0x8528,
        0x9509,
        0xE5EE,
        0xF5CF,
        0xC5AC,
        0xD58D,
        0x3653,
        0x2672,
        0x1611,
        0x0630,
        0x76D7,
        0x66F6,
        0x5695,
        0x46B4,
        0xB75B,
        0xA77A,
        0x9719,
        0x8738,
        0xF7DF,
        0xE7FE,
        0xD79D,
        0xC7BC,
        0x48C4,
        0x58E5,
        0x6886,
        0x78A7,
        0x0840,
        0x1861,
        0x2802,
        0x3823,
        0xC9CC,
        0xD9ED,
        0xE98E,
        0xF9AF,
        0x8948,
        0x9969,
        0xA90A,
        0xB92B,
        0x5AF5,
        0x4AD4,
        0x7AB7,
        0x6A96,
        0x1A71,
        0x0A50,
        0x3A33,
        0x2A12,
        0xDBFD,
        0xCBDC,
        0xFBBF,
        0xEB9E,
        0x9B79,
        0x8B58,
        0xBB3B,
        0xAB1A,
        0x6CA6,
        0x7C87,
        0x4CE4,
        0x5CC5,
        0x2C22,
        0x3C03,
        0x0C60,
        0x1C41,
        0xEDAE,
        0xFD8F,
        0xCDEC,
        0xDDCD,
        0xAD2A,
        0xBD0B,
        0x8D68,
        0x9D49,
        0x7E97,
        0x6EB6,
        0x5ED5,
        0x4EF4,
        0x3E13,
        0x2E32,
        0x1E51,
        0x0E70,
        0xFF9F,
        0xEFBE,
        0xDFDD,
        0xCFFC,
        0xBF1B,
        0xAF3A,
        0x9F59,
        0x8F78,
        0x9188,
        0x81A9,
        0xB1CA,
        0xA1EB,
        0xD10C,
        0xC12D,
        0xF14E,
        0xE16F,
        0x1080,
        0x00A1,
        0x30C2,
        0x20E3,
        0x5004,
        0x4025,
        0x7046,
        0x6067,
        0x83B9,
        0x9398,
        0xA3FB,
        0xB3DA,
        0xC33D,
        0xD31C,
        0xE37F,
        0xF35E,
        0x02B1,
        0x1290,
        0x22F3,
        0x32D2,
        0x4235,
        0x5214,
        0x6277,
        0x7256,
        0xB5EA,
        0xA5CB,
        0x95A8,
        0x8589,
        0xF56E,
        0xE54F,
        0xD52C,
        0xC50D,
        0x34E2,
        0x24C3,
        0x14A0,
        0x0481,
        0x7466,
        0x6447,
        0x5424,
        0x4405,
        0xA7DB,
        0xB7FA,
        0x8799,
        0x97B8,
        0xE75F,
        0xF77E,
        0xC71D,
        0xD73C,
        0x26D3,
        0x36F2,
        0x0691,
        0x16B0,
        0x6657,
        0x7676,
        0x4615,
        0x5634,
        0xD94C,
        0xC96D,
        0xF90E,
        0xE92F,
        0x99C8,
        0x89E9,
        0xB98A,
        0xA9AB,
        0x5844,
        0x4865,
        0x7806,
        0x6827,
        0x18C0,
        0x08E1,
        0x3882,
        0x28A3,
        0xCB7D,
        0xDB5C,
        0xEB3F,
        0xFB1E,
        0x8BF9,
        0x9BD8,
        0xABBB,
        0xBB9A,
        0x4A75,
        0x5A54,
        0x6A37,
        0x7A16,
        0x0AF1,
        0x1AD0,
        0x2AB3,
        0x3A92,
        0xFD2E,
        0xED0F,
        0xDD6C,
        0xCD4D,
        0xBDAA,
        0xAD8B,
        0x9DE8,
        0x8DC9,
        0x7C26,
        0x6C07,
        0x5C64,
        0x4C45,
        0x3CA2,
        0x2C83,
        0x1CE0,
        0x0CC1,
        0xEF1F,
        0xFF3E,
        0xCF5D,
        0xDF7C,
        0xAF9B,
        0xBFBA,
        0x8FD9,
        0x9FF8,
        0x6E17,
        0x7E36,
        0x4E55,
        0x5E74,
        0x2E93,
        0x3EB2,
        0x0ED1,
        0x1EF0,
    ]

    def crc16(data: bytes) -> int:
        crc = 0

        for byte in data:
            crc = ((crc << 8) & 0xFF00) ^ x_mode_m_crc16_lookup[
                ((crc >> 8) & 0xFF) ^ byte
            ]

        return crc & 0xFFFF

    def hash_slot(key: bytes) -> int:
        start = key.find(b"{")

        if start > -1:
            end = key.find(b"}", start + 1)

            if end > -1 and end != start + 1:
                key = key[start + 1 : end]

        return crc16(key) % 16384


__all__ = [
    "hash_slot",
    "EncodingInsensitiveDict",
    "CaseAndEncodingInsensitiveEnum",
]



from __future__ import annotations

from datetime import timedelta

from ..commands._utils import normalized_milliseconds
from ..commands._wrappers import ClusterCommandConfig
from ..commands.constants import CommandGroup, CommandName, NodeFlag
from ..response._callbacks import (
    ClusterEnsureConsistent,
    ClusterMergeSets,
    ListCallback,
    SetCallback,
    SimpleStringCallback,
)
from ..tokens import PrefixToken, PureToken
from ..typing import (
    AnyStr,
    CommandArgList,
    Dict,
    KeyT,
    List,
    Literal,
    Optional,
    Parameters,
    ResponsePrimitive,
    Set,
    StringT,
    Tuple,
    Union,
    ValueT,
)
from .base import Module, ModuleGroup, module_command
from .response._callbacks.graph import (
    ConfigGetCallback,
    GraphSlowLogCallback,
    QueryCallback,
)
from .response.types import GraphQueryResult, GraphSlowLogInfo


class RedisGraph(Module[AnyStr]):
    NAME = "graph"
    FULL_NAME = "RedisGraph"
    DESCRIPTION = """RedisGraph is a queryable Property Graph database that uses sparse matrices 
to represent the adjacency matrix in graphs and linear algebra to query the graph.
    """
    DOCUMENTATION_URL = "https://redis.io/docs/stack/graph/"


class Graph(ModuleGroup[AnyStr]):
    MODULE = RedisGraph
    COMMAND_GROUP = CommandGroup.GRAPH

    @module_command(
        CommandName.GRAPH_QUERY,
        module=MODULE,
        version_introduced="1.0.0",
        group=COMMAND_GROUP,
    )
    async def query(
        self,
        graph: KeyT,
        query: StringT,
        timeout: Optional[Union[int, timedelta]] = None,
    ) -> GraphQueryResult[AnyStr]:
        """
        Executes the given query against a specified graph

        :param graph: The name of the graph to query.
        :param query: The query to execute.
        :param timeout: The maximum amount of time (milliseconds) to wait for the query to complete
        :return: The result set of the executed query.
        """
        pieces: CommandArgList = [graph, query]

        if timeout is not None:
            pieces.extend([PrefixToken.TIMEOUT, normalized_milliseconds(timeout)])

        pieces.append(b"--compact")
        return await self.execute_module_command(
            CommandName.GRAPH_QUERY,
            *pieces,
            callback=QueryCallback[AnyStr](graph),
            query=query,
        )

    @module_command(
        CommandName.GRAPH_RO_QUERY,
        module=MODULE,
        version_introduced="2.2.8",
        group=COMMAND_GROUP,
    )
    async def ro_query(
        self,
        graph: KeyT,
        query: StringT,
        timeout: Optional[Union[int, timedelta]] = None,
    ) -> GraphQueryResult[AnyStr]:
        """
        Executes a given read only query against a specified graph

        :param graph: The name of the graph to query.
        :param query: The query to execute.
        :param timeout: The maximum amount of time (milliseconds) to wait for the query to complete.
        :return: The result set for the read-only query or an error if a write query was given.
        """
        pieces: CommandArgList = [graph, query]
        if timeout is not None:
            pieces.extend([PrefixToken.TIMEOUT, normalized_milliseconds(timeout)])
        pieces.append(b"--compact")

        return await self.execute_module_command(
            CommandName.GRAPH_RO_QUERY,
            *pieces,
            callback=QueryCallback[AnyStr](graph),
            query=query,
        )

    @module_command(
        CommandName.GRAPH_DELETE,
        module=MODULE,
        version_introduced="1.0.0",
        group=COMMAND_GROUP,
    )
    async def delete(self, graph: KeyT) -> bool:
        """
        Completely removes the graph and all of its entities

        Deletes the entire graph and all of its entities.

        :param graph: The name of the graph to be deleted.
        """

        return await self.execute_module_command(
            CommandName.GRAPH_DELETE,
            graph,
            callback=SimpleStringCallback(
                prefix_match=True, ok_values={"Graph removed"}
            ),
        )

    @module_command(
        CommandName.GRAPH_EXPLAIN,
        module=MODULE,
        version_introduced="2.0.0",
        group=COMMAND_GROUP,
    )
    async def explain(self, graph: KeyT, query: StringT) -> List[AnyStr]:
        """

        Constructs a query execution plan for the given :paramref:`graph` and
        :paramref:`query`, but does not execute it.

        :param graph: The name of the graph to execute the query on.
        :param query: The query to construct the execution plan for.

        :return: A list of strings representing the query execution plan.
        """

        return await self.execute_module_command(
            CommandName.GRAPH_EXPLAIN, graph, query, callback=ListCallback[AnyStr]()
        )

    @module_command(
        CommandName.GRAPH_PROFILE,
        module=MODULE,
        version_introduced="2.0.0",
        group=COMMAND_GROUP,
    )
    async def profile(
        self,
        graph: KeyT,
        query: StringT,
        timeout: Optional[Union[int, timedelta]] = None,
    ) -> List[AnyStr]:
        """
        Executes a query and returns an execution plan augmented with metrics for each
        operation's execution

        :param graph: The name of the graph to execute the query on.
        :param query: The query to execute and return a profile for.
        :param timeout: Optional timeout for the query execution in milliseconds.
        :return: A string representation of a query execution plan, with details on results produced
         by and time spent in each operation.
        """
        pieces: CommandArgList = [graph, query]
        if timeout is not None:
            pieces.extend([PrefixToken.TIMEOUT, normalized_milliseconds(timeout)])
        return await self.execute_module_command(
            CommandName.GRAPH_PROFILE, *pieces, callback=ListCallback[AnyStr]()
        )

    @module_command(
        CommandName.GRAPH_SLOWLOG,
        module=MODULE,
        version_introduced="2.0.12",
        arguments={"reset": {"version_introduced": "2.12.0"}},
        group=COMMAND_GROUP,
    )
    async def slowlog(
        self, graph: KeyT, reset: bool = False
    ) -> Union[Tuple[GraphSlowLogInfo, ...], bool]:
        """
        Returns a list containing up to 10 of the slowest queries issued against the given graph

        :param graph: The name of the graph
        :param reset: If ``True``, the slowlog will be reset
        :return: The slowlog for the given graph or ``True`` if the slowlog was reset
        """
        pieces: CommandArgList = [graph]
        if reset:
            pieces.append(PureToken.RESET)
            return await self.execute_module_command(
                CommandName.GRAPH_SLOWLOG,
                *pieces,
                callback=SimpleStringCallback(),
            )
        else:
            return await self.execute_module_command(
                CommandName.GRAPH_SLOWLOG,
                *pieces,
                callback=GraphSlowLogCallback(),
            )

    @module_command(
        CommandName.GRAPH_CONFIG_GET,
        module=MODULE,
        version_introduced="2.2.11",
        group=COMMAND_GROUP,
        cluster=ClusterCommandConfig(
            route=NodeFlag.RANDOM,
        ),
    )
    async def config_get(
        self, name: StringT
    ) -> Union[Dict[AnyStr, ResponsePrimitive], ResponsePrimitive]:
        """
        Retrieves a RedisGraph configuration

        :param name: The name of the configuration parameter to retrieve.
        :return: The value of the configuration parameter. If :paramref:`name`
         is ``*``, a mapping of all configuration parameters to their values
        """
        return await self.execute_module_command(
            CommandName.GRAPH_CONFIG_GET,
            name,
            callback=ConfigGetCallback[AnyStr](),
        )

    @module_command(
        CommandName.GRAPH_CONFIG_SET,
        module=MODULE,
        version_introduced="2.2.11",
        group=COMMAND_GROUP,
        cluster=ClusterCommandConfig(
            route=NodeFlag.PRIMARIES,
            combine=ClusterEnsureConsistent[AnyStr](),
        ),
    )
    async def config_set(self, name: StringT, value: ValueT) -> bool:
        """
        Updates a RedisGraph configuration

        :param name: The name of the configuration parameter to set.
        :param value: The value to set the configuration parameter to.
        :return: True if the configuration parameter was set successfully, False otherwise.
        """
        return await self.execute_module_command(
            CommandName.GRAPH_CONFIG_SET, name, value, callback=SimpleStringCallback()
        )

    @module_command(
        CommandName.GRAPH_LIST,
        module=MODULE,
        version_introduced="2.4.3",
        group=COMMAND_GROUP,
        cluster=ClusterCommandConfig(
            route=NodeFlag.PRIMARIES,
            combine=ClusterMergeSets[AnyStr](),
        ),
    )
    async def list(self) -> Set[AnyStr]:
        """
        Lists all graph keys in the keyspace

        :return: A list of graph keys in the keyspace.
        """

        return await self.execute_module_command(
            CommandName.GRAPH_LIST, callback=SetCallback[AnyStr]()
        )

    @module_command(
        CommandName.GRAPH_CONSTRAINT_DROP,
        module=MODULE,
        version_introduced="2.12.0",
        group=COMMAND_GROUP,
    )
    async def constraint_drop(
        self,
        graph: KeyT,
        type: Literal[PureToken.MANDATORY, PureToken.UNIQUE],
        node: Optional[StringT] = None,
        relationship: Optional[StringT] = None,
        properties: Optional[Parameters[StringT]] = None,
    ) -> bool:
        """
        Deletes a constraint from specified graph

        :param graph: The name of the RedisGraph.
        :param type: The type of constraint to drop.
        :param node: The name of the node to drop the constraint from
        :param relationship: The name of the relationship to drop the constraint from
        :param properties: The properties to drop the constraint from

        :return: True if the constraint was successfully dropped, False otherwise.
        """
        pieces: CommandArgList = [graph, type]
        if node is not None:
            pieces.extend([PrefixToken.NODE, node])
        if relationship is not None:
            pieces.extend([PrefixToken.RELATIONSHIP, relationship])
        if properties:
            _props: List[StringT] = list(properties)
            pieces.extend([PrefixToken.PROPERTIES, len(_props), *_props])

        return await self.execute_module_command(
            CommandName.GRAPH_CONSTRAINT_DROP,
            *pieces,
            callback=SimpleStringCallback(),
        )

    @module_command(
        CommandName.GRAPH_CONSTRAINT_CREATE,
        module=MODULE,
        version_introduced="2.12.0",
        group=COMMAND_GROUP,
    )
    async def constraint_create(
        self,
        graph: KeyT,
        type: Literal[PureToken.MANDATORY, PureToken.UNIQUE],
        node: Optional[StringT] = None,
        relationship: Optional[StringT] = None,
        properties: Optional[Parameters[StringT]] = None,
    ) -> bool:
        """
        Creates a constraint on specified graph


        :param graph: The name of the graph.
        :param type: The type of constraint to create.
        :param node: The label of the node to apply the constraint to.
        :param relationship: The type of relationship to apply the constraint to.
        :param properties: The properties to apply the constraint to.
        :return: True if the constraint was created successfully, False otherwise.
        """
        pieces: CommandArgList = [graph, type]

        if node is not None:
            pieces.extend([PrefixToken.NODE, node])
        if relationship is not None:
            pieces.extend([PrefixToken.RELATIONSHIP, relationship])
        if properties:
            _props: List[StringT] = list(properties)
            pieces.extend([PrefixToken.PROPERTIES, len(_props), *_props])
        return await self.execute_module_command(
            CommandName.GRAPH_CONSTRAINT_CREATE,
            *pieces,
            callback=SimpleStringCallback(ok_values={"PENDING"}),
        )



from __future__ import annotations

from deprecated.sphinx import versionadded

from .._utils import dict_to_flat_list
from ..commands._validators import mutually_inclusive_parameters
from ..commands._wrappers import CacheConfig
from ..commands.constants import CommandFlag, CommandGroup, CommandName
from ..response._callbacks import (
    BoolCallback,
    BoolsCallback,
    DictCallback,
    FirstValueCallback,
    FloatCallback,
    FloatsCallback,
    IntCallback,
    MixedTupleCallback,
    SimpleStringCallback,
    TupleCallback,
)
from ..tokens import PrefixToken, PureToken
from ..typing import (
    AnyStr,
    CommandArgList,
    Dict,
    KeyT,
    List,
    Literal,
    Mapping,
    Optional,
    Parameters,
    ResponsePrimitive,
    StringT,
    Tuple,
    Union,
    ValueT,
)
from .base import Module, ModuleGroup, module_command


class RedisBloom(Module[AnyStr]):
    NAME = "bf"
    FULL_NAME = "RedisBloom"
    DESCRIPTION = """RedisBloom is a Redis module that implements various probabilistic
data structures such as BloomFilter.
    """
    DOCUMENTATION_URL = "https://redis.io/docs/stack/bloom/"


@versionadded(version="4.12")
class BloomFilter(ModuleGroup[AnyStr]):
    MODULE = RedisBloom
    COMMAND_GROUP = CommandGroup.BF

    @module_command(
        CommandName.BF_RESERVE,
        group=COMMAND_GROUP,
        version_introduced="1.0.0",
        module=MODULE,
    )
    async def reserve(
        self,
        key: KeyT,
        error_rate: Union[int, float],
        capacity: int,
        expansion: Optional[int] = None,
        nonscaling: Optional[bool] = None,
    ) -> bool:
        """
        Creates a new Bloom Filter

        :param key: The key under which the filter is found.
        :param error_rate: The desired probability for false positives.
        :param capacity: The number of entries intended to be added to the filter.
        :param expansion: The size of the new sub-filter when `capacity` is reached.
        :param nonscaling: Prevents the filter from creating additional sub-filters.
        """
        pieces: CommandArgList = [key, error_rate, capacity]
        if expansion is not None:
            pieces.extend([PrefixToken.EXPANSION, expansion])
        if nonscaling:
            pieces.append(PureToken.NONSCALING)

        return await self.execute_module_command(
            CommandName.BF_RESERVE, *pieces, callback=SimpleStringCallback()
        )

    @module_command(
        CommandName.BF_ADD,
        group=COMMAND_GROUP,
        version_introduced="1.0.0",
        module=MODULE,
    )
    async def add(self, key: KeyT, item: ValueT) -> bool:
        """
        Adds an item to a Bloom Filter

        :param key: The key under which the filter is found.
        :param item: The item to add to the filter.
        """
        pieces: CommandArgList = [key, item]

        return await self.execute_module_command(
            CommandName.BF_ADD, *pieces, callback=BoolCallback()
        )

    @module_command(
        CommandName.BF_MADD,
        group=COMMAND_GROUP,
        version_introduced="1.0.0",
        module=MODULE,
    )
    async def madd(self, key: KeyT, items: Parameters[ValueT]) -> Tuple[bool, ...]:
        """
        Adds one or more items to a Bloom Filter. A filter will be created if it does not exist

        :param key: The key under which the filter is found.
        :param items: One or more items to add.
        """
        pieces: CommandArgList = [key, *items]

        return await self.execute_module_command(
            CommandName.BF_MADD, *pieces, callback=BoolsCallback()
        )

    @module_command(
        CommandName.BF_INSERT,
        group=COMMAND_GROUP,
        version_introduced="1.0.0",
        module=MODULE,
    )
    async def insert(
        self,
        key: KeyT,
        items: Parameters[ValueT],
        capacity: Optional[int] = None,
        error: Optional[Union[int, float]] = None,
        expansion: Optional[int] = None,
        nocreate: Optional[bool] = None,
        nonscaling: Optional[bool] = None,
    ) -> Tuple[bool, ...]:
        """
        Adds one or more items to a Bloom Filter. A filter will be created if it
        does not exist

        :param key: The key under which the filter is found.
        :param items: One or more items to add.
        :param capacity: The desired capacity for the filter to be created.
        :param error: The error ratio of the newly created filter if it does not yet exist.
        :param expansion: The expansion factor for the filter when capacity is reached.
        :param nocreate: Indicates that the filter should not be created if it does not
         already exist.
        :param nonscaling: Prevents the filter from creating additional sub-filters
         if initial capacity is reached.
        """
        pieces: CommandArgList = [key]
        if capacity is not None:
            pieces.extend([PrefixToken.CAPACITY, capacity])
        if error is not None:
            pieces.extend([PrefixToken.ERROR, error])
        if expansion is not None:
            pieces.extend([PrefixToken.EXPANSION, expansion])
        if nocreate:
            pieces.append(PureToken.NOCREATE)
        if nonscaling:
            pieces.append(PureToken.NONSCALING)
        pieces.append(PureToken.ITEMS)
        pieces.extend(items)
        return await self.execute_module_command(
            CommandName.BF_INSERT, *pieces, callback=BoolsCallback()
        )

    @module_command(
        CommandName.BF_EXISTS,
        group=COMMAND_GROUP,
        version_introduced="1.0.0",
        module=MODULE,
        flags={CommandFlag.READONLY},
        cache_config=CacheConfig(lambda *a, **_: a[0]),
    )
    async def exists(self, key: KeyT, item: ValueT) -> bool:
        """
        Checks whether an item exists in a Bloom Filter

        :param key: The key under which the filter is found.
        :param item: The item to check for existence.
        """
        return await self.execute_module_command(
            CommandName.BF_EXISTS, key, item, callback=BoolCallback()
        )

    @module_command(
        CommandName.BF_MEXISTS,
        group=COMMAND_GROUP,
        version_introduced="1.0.0",
        module=MODULE,
        flags={CommandFlag.READONLY},
        cache_config=CacheConfig(lambda *a, **_: a[0]),
    )
    async def mexists(self, key: KeyT, items: Parameters[ValueT]) -> Tuple[bool, ...]:
        """
        Checks whether one or more items exist in a Bloom Filter

        :param key: The key under which the filter is found.
        :param items: One or more items to check.
        """
        return await self.execute_module_command(
            CommandName.BF_MEXISTS, key, *items, callback=BoolsCallback()
        )

    @module_command(
        CommandName.BF_SCANDUMP,
        group=COMMAND_GROUP,
        version_introduced="1.0.0",
        module=MODULE,
    )
    async def scandump(self, key: KeyT, iterator: int) -> Tuple[int, Optional[bytes]]:
        """
        Begins an incremental save of the bloom filter

        :param key: The key under which the filter is found.
        :param iterator: Iterator value; either 0 to start a dump or the iterator
         from a previous invocation of this command.

        :return: A tuple of iterator value and data. If iterator is 0, iteration has
         completed. The iterator-data pair should be passed to :meth:`loadchunk` when
         restoring the filter.
        """

        return await self.execute_module_command(
            CommandName.BF_SCANDUMP,
            key,
            iterator,
            callback=MixedTupleCallback[int, Optional[bytes]](),
            decode=False,
        )

    @module_command(
        CommandName.BF_LOADCHUNK,
        group=COMMAND_GROUP,
        version_introduced="1.0.0",
        module=MODULE,
    )
    async def loadchunk(self, key: KeyT, iterator: int, data: bytes) -> bool:
        """
        Restores a filter previously saved using :meth:`scandump`

        :param key: Name of the key to restore.
        :param iterator: Iterator value associated with the data chunk.
        :param data: Current data chunk.
        """
        pieces: CommandArgList = [key, iterator, data]

        return await self.execute_module_command(
            CommandName.BF_LOADCHUNK, *pieces, callback=SimpleStringCallback()
        )

    @module_command(
        CommandName.BF_INFO,
        group=COMMAND_GROUP,
        version_introduced="1.0.0",
        module=MODULE,
        flags={CommandFlag.READONLY},
        cache_config=CacheConfig(lambda *a, **_: a[0]),
    )
    async def info(
        self,
        key: KeyT,
        single_value: Optional[
            Literal[
                PureToken.CAPACITY,
                PureToken.EXPANSION,
                PureToken.FILTERS,
                PureToken.ITEMS,
                PureToken.SIZE,
            ]
        ] = None,
    ) -> Union[Dict[AnyStr, int], int]:
        """
        Returns information about a Bloom Filter

        :param key: The key name for an existing Bloom filter.
        :param single_value: Optional parameter to get a specific information field.

        :return: A dictionary with all information fields if :paramref:`single_value`
         is not specified, or an integer representing the value of the specified field.
        """
        if single_value:
            return await self.execute_module_command(
                CommandName.BF_INFO,
                key,
                single_value,
                callback=FirstValueCallback[int](),
            )
        else:
            return await self.execute_module_command(
                CommandName.BF_INFO, key, callback=DictCallback[AnyStr, int]()
            )

    @module_command(
        CommandName.BF_CARD,
        group=COMMAND_GROUP,
        version_introduced="2.4.4",
        module=MODULE,
        flags={CommandFlag.READONLY},
        cache_config=CacheConfig(lambda *a, **_: a[0]),
    )
    async def card(self, key: KeyT) -> int:
        """
        Returns the cardinality of a Bloom filter

        :param key: The key name for an existing Bloom filter.
        """
        return await self.execute_module_command(
            CommandName.BF_CARD, key, callback=IntCallback()
        )


@versionadded(version="4.12")
class CuckooFilter(ModuleGroup[AnyStr]):
    MODULE = RedisBloom
    COMMAND_GROUP = CommandGroup.CF

    @module_command(
        CommandName.CF_RESERVE,
        group=COMMAND_GROUP,
        version_introduced="1.0.0",
        module=MODULE,
    )
    async def reserve(
        self,
        key: KeyT,
        capacity: int,
        bucketsize: Optional[int] = None,
        maxiterations: Optional[int] = None,
        expansion: Optional[int] = None,
    ) -> bool:
        """
        Creates a new Cuckoo Filter

        :param key: The name of the filter.
        :param capacity: Estimated capacity for the filter.
        :param bucketsize: Number of items in each bucket.
        :param maxiterations: Number of attempts to swap items between buckets
         before declaring filter as full and creating an additional filter.
        :param expansion: When a new filter is created, its size is the size of the
         current filter multiplied by ``expansion``.
        """
        pieces: CommandArgList = [key, capacity]
        if bucketsize is not None:
            pieces.extend([PrefixToken.BUCKETSIZE, bucketsize])
        if maxiterations is not None:
            pieces.extend([PrefixToken.MAXITERATIONS, maxiterations])
        if expansion is not None:
            pieces.extend([PrefixToken.EXPANSION, expansion])
        return await self.execute_module_command(
            CommandName.CF_RESERVE, *pieces, callback=SimpleStringCallback()
        )

    @module_command(
        CommandName.CF_ADD,
        group=COMMAND_GROUP,
        version_introduced="1.0.0",
        module=MODULE,
    )
    async def add(self, key: KeyT, item: ValueT) -> bool:
        """
        Adds an item to a Cuckoo Filter

        :param key: The name of the filter.
        :param item: The item to add.
        """
        pieces: CommandArgList = [key, item]

        return await self.execute_module_command(
            CommandName.CF_ADD, *pieces, callback=BoolCallback()
        )

    @module_command(
        CommandName.CF_ADDNX,
        group=COMMAND_GROUP,
        version_introduced="1.0.0",
        module=MODULE,
    )
    async def addnx(self, key: KeyT, item: ValueT) -> bool:
        """
        Adds an item to a Cuckoo Filter if the item did not exist previously.

        :param key: The name of the filter.
        :param item: The item to add.
        """
        pieces: CommandArgList = [key, item]

        return await self.execute_module_command(
            CommandName.CF_ADDNX, *pieces, callback=BoolCallback()
        )

    @module_command(
        CommandName.CF_INSERT,
        group=COMMAND_GROUP,
        version_introduced="1.0.0",
        module=MODULE,
    )
    async def insert(
        self,
        key: KeyT,
        items: Parameters[ValueT],
        capacity: Optional[int] = None,
        nocreate: Optional[bool] = None,
    ) -> Tuple[bool, ...]:
        """
        Adds one or more items to a Cuckoo Filter. A filter will be created if it does not exist

        :param key: The name of the filter.
        :param items: One or more items to add.
        :param capacity: Specifies the desired capacity of the new filter, if this filter
         does not exist yet.
        :param nocreate: If specified, prevents automatic filter creation if the filter
         does not exist.
        :return: A tuple of boolean values indicating if the command was executed correctly.
        """
        pieces: CommandArgList = [key]
        if capacity is not None:
            pieces.extend([PrefixToken.CAPACITY, capacity])
        if nocreate is not None:
            pieces.append(PureToken.NOCREATE)
        pieces.append(PureToken.ITEMS)
        pieces.extend(items)

        return await self.execute_module_command(
            CommandName.CF_INSERT, *pieces, callback=BoolsCallback()
        )

    @module_command(
        CommandName.CF_INSERTNX,
        group=COMMAND_GROUP,
        version_introduced="1.0.0",
        module=MODULE,
    )
    async def insertnx(
        self,
        key: KeyT,
        items: Parameters[ValueT],
        capacity: Optional[int] = None,
        nocreate: Optional[bool] = None,
    ) -> Tuple[bool, ...]:
        """
        Adds one or more items to a Cuckoo Filter if the items did not exist previously.
        A filter will be created if it does not exist

        :param key: The name of the filter.
        :param items: One or more items to add.
        :param capacity: Specifies the desired capacity of the new filter,
         if this filter does not exist yet.
        :param nocreate: If specified, prevents automatic filter creation
         if the filter does not exist.
        """
        pieces: CommandArgList = [key]
        if capacity is not None:
            pieces.extend([PrefixToken.CAPACITY, capacity])
        if nocreate is not None:
            pieces.append(PureToken.NOCREATE)
        pieces.append(PureToken.ITEMS)
        pieces.extend(items)

        return await self.execute_module_command(
            CommandName.CF_INSERTNX, *pieces, callback=BoolsCallback()
        )

    @module_command(
        CommandName.CF_EXISTS,
        group=COMMAND_GROUP,
        version_introduced="1.0.0",
        module=MODULE,
        flags={CommandFlag.READONLY},
        cache_config=CacheConfig(lambda *a, **_: a[0]),
    )
    async def exists(self, key: KeyT, item: ValueT) -> bool:
        """
        Checks whether an item exist in a Cuckoo Filter

        :param key: The name of the filter.
        :param item: The item to check for.
        """
        pieces: CommandArgList = [key, item]

        return await self.execute_module_command(
            CommandName.CF_EXISTS, *pieces, callback=BoolCallback()
        )

    @module_command(
        CommandName.CF_MEXISTS,
        group=COMMAND_GROUP,
        version_introduced="1.0.0",
        module=MODULE,
        flags={CommandFlag.READONLY},
        cache_config=CacheConfig(lambda *a, **_: a[0]),
    )
    async def mexists(self, key: KeyT, items: Parameters[ValueT]) -> Tuple[bool, ...]:
        """
        Checks whether one or more items exist in a Cuckoo Filter

        :param key: The name of the filter.
        :param items: The item(s) to check for.
        """
        pieces: CommandArgList = [key, *items]

        return await self.execute_module_command(
            CommandName.CF_MEXISTS, *pieces, callback=BoolsCallback()
        )

    @module_command(
        CommandName.CF_DEL,
        group=COMMAND_GROUP,
        version_introduced="1.0.0",
        module=MODULE,
    )
    async def delete(self, key: KeyT, item: ValueT) -> bool:
        """
        Deletes an item from a Cuckoo Filter

        :param key: The name of the filter.
        :param item: The item to delete from the filter.
        """
        pieces: CommandArgList = [key, item]

        return await self.execute_module_command(
            CommandName.CF_DEL, *pieces, callback=BoolCallback()
        )

    @module_command(
        CommandName.CF_COUNT,
        group=COMMAND_GROUP,
        version_introduced="1.0.0",
        module=MODULE,
        flags={CommandFlag.READONLY},
        cache_config=CacheConfig(lambda *a, **_: a[0]),
    )
    async def count(self, key: KeyT, item: ValueT) -> int:
        """
        Return the number of times an item might be in a Cuckoo Filter

        :param key: The name of the filter.
        :param item: The item to count.
        """
        pieces: CommandArgList = [key, item]

        return await self.execute_module_command(
            CommandName.CF_COUNT, *pieces, callback=IntCallback()
        )

    @module_command(
        CommandName.CF_SCANDUMP,
        group=COMMAND_GROUP,
        version_introduced="1.0.0",
        module=MODULE,
    )
    async def scandump(self, key: KeyT, iterator: int) -> Tuple[int, Optional[bytes]]:
        """
        Begins an incremental save of the bloom filter

        :param key: Name of the filter.
        :param iterator: Iterator value. This is either 0, or the iterator from a
         previous invocation of this command.
        """
        pieces: CommandArgList = [key, iterator]

        return await self.execute_module_command(
            CommandName.CF_SCANDUMP,
            *pieces,
            decode=False,
            callback=MixedTupleCallback[int, Optional[bytes]](),
        )

    @module_command(
        CommandName.CF_LOADCHUNK,
        group=COMMAND_GROUP,
        version_introduced="1.0.0",
        module=MODULE,
    )
    async def loadchunk(self, key: KeyT, iterator: int, data: StringT) -> bool:
        """
        Restores a filter previously saved using SCANDUMP

        :param key: Name of the key to restore.
        :param iter: Iterator value associated with :paramref:`data` (returned by :meth:`scandump`).
        :param data: Current data chunk (returned by :meth:`scandump`).

        """
        pieces: CommandArgList = [key, iterator, data]

        return await self.execute_module_command(
            CommandName.CF_LOADCHUNK, *pieces, callback=SimpleStringCallback()
        )

    @module_command(
        CommandName.CF_INFO,
        group=COMMAND_GROUP,
        version_introduced="1.0.0",
        module=MODULE,
        flags={CommandFlag.READONLY},
    )
    async def info(self, key: KeyT) -> Dict[AnyStr, ResponsePrimitive]:
        """
        Returns information about a Cuckoo Filter

        :param key: The name of the filter.
        """

        return await self.execute_module_command(
            CommandName.CF_INFO, key, callback=DictCallback[AnyStr, ResponsePrimitive]()
        )


@versionadded(version="4.12")
class CountMinSketch(ModuleGroup[AnyStr]):
    MODULE = RedisBloom
    COMMAND_GROUP = CommandGroup.CMS

    @module_command(
        CommandName.CMS_INITBYDIM,
        group=COMMAND_GROUP,
        version_introduced="2.0.0",
        module=MODULE,
    )
    async def initbydim(self, key: KeyT, width: int, depth: int) -> bool:
        """
        Initializes a Count-Min Sketch to dimensions specified by user

        :param key: Name of the sketch.
        :param width: Number of counters in each array. Reduces error size.
        :param depth: Number of counter-arrays. Reduces error probability.
        """
        return await self.execute_module_command(
            CommandName.CMS_INITBYDIM,
            key,
            width,
            depth,
            callback=SimpleStringCallback(),
        )

    @module_command(
        CommandName.CMS_INITBYPROB,
        group=COMMAND_GROUP,
        version_introduced="2.0.0",
        module=MODULE,
    )
    async def initbyprob(
        self, key: KeyT, error: Union[int, float], probability: Union[int, float]
    ) -> bool:
        """
        Initializes a Count-Min Sketch to accommodate requested tolerances.

        :param key: Name of the sketch.
        :param error: Estimate size of error as a percent of total counted items.
        :param probability: Desired probability for inflated count as a decimal value
         between 0 and 1.
        """
        return await self.execute_module_command(
            CommandName.CMS_INITBYPROB,
            key,
            error,
            probability,
            callback=SimpleStringCallback(),
        )

    @module_command(
        CommandName.CMS_INCRBY,
        group=COMMAND_GROUP,
        version_introduced="2.0.0",
        module=MODULE,
    )
    async def incrby(self, key: KeyT, items: Mapping[AnyStr, int]) -> Tuple[int, ...]:
        """
        Increases the count of one or more items by increment

        :param key: The name of the HyperLogLog sketch.
        :param items: A dictionary containing the items to increment and
         their respective increments.
        """

        return await self.execute_module_command(
            CommandName.CMS_INCRBY,
            key,
            *dict_to_flat_list(items),
            callback=TupleCallback[int](),
        )

    @module_command(
        CommandName.CMS_QUERY,
        group=COMMAND_GROUP,
        version_introduced="2.0.0",
        module=MODULE,
        flags={CommandFlag.READONLY},
        cache_config=CacheConfig(lambda *a, **_: a[0]),
    )
    async def query(
        self,
        key: KeyT,
        items: Parameters[StringT],
    ) -> Tuple[int, ...]:
        """
        Returns the count for one or more items in a sketch

        :param key: The name of the Count-Min Sketch.
        :param items: One or more items for which to return the count.
        """
        pieces: CommandArgList = [key, *items]

        return await self.execute_module_command(
            CommandName.CMS_QUERY, *pieces, callback=TupleCallback[int]()
        )

    @module_command(
        CommandName.CMS_MERGE,
        group=COMMAND_GROUP,
        version_introduced="2.0.0",
        module=MODULE,
    )
    async def merge(
        self,
        destination: KeyT,
        sources: Parameters[KeyT],
        weights: Optional[Parameters[Union[int, float]]] = None,
    ) -> bool:
        """
        Merges several sketches into one sketch

        :param destination: The name of the destination sketch. Must be initialized.
        :param sources: Names of the source sketches to be merged.
        :param weights: Multiples of each sketch. Default is 1.
        """
        _sources: List[KeyT] = list(sources)
        pieces: CommandArgList = [destination, len(_sources), *_sources]
        if weights:
            pieces.append(PrefixToken.WEIGHTS)
            pieces.extend(weights)

        return await self.execute_module_command(
            CommandName.CMS_MERGE, *pieces, callback=SimpleStringCallback()
        )

    @module_command(
        CommandName.CMS_INFO,
        group=COMMAND_GROUP,
        version_introduced="2.0.0",
        module=MODULE,
        flags={CommandFlag.READONLY},
        cache_config=CacheConfig(lambda *a, **_: a[0]),
    )
    async def info(self, key: KeyT) -> Dict[AnyStr, int]:
        """
        Returns information about a sketch

        :param key: The name of the sketch.
        :return: A dictionary containing the width, depth, and total count of the sketch.
        """

        return await self.execute_module_command(
            CommandName.CMS_INFO,
            key,
            callback=DictCallback[AnyStr, int](),
        )


@versionadded(version="4.12")
class TopK(ModuleGroup[AnyStr]):
    MODULE = RedisBloom
    COMMAND_GROUP = CommandGroup.TOPK

    @mutually_inclusive_parameters("width", "depth", "decay")
    @module_command(
        CommandName.TOPK_RESERVE,
        group=COMMAND_GROUP,
        version_introduced="2.0.0",
        module=MODULE,
    )
    async def reserve(
        self,
        key: KeyT,
        topk: int,
        width: Optional[int] = None,
        depth: Optional[int] = None,
        decay: Optional[Union[int, float]] = None,
    ) -> bool:
        """
        Reserve a TopK sketch with specified parameters.

        :param key: Name of the TOP-K sketch.
        :param topk: Number of top occurring items to keep.
        :param width: Number of counters kept in each array.
        :param depth: Number of arrays.
        :param decay: The probability of reducing a counter in an occupied bucket.
         It is raised to power of it's counter (``decay ^ bucket[i].counter``).
         Therefore, as the counter gets higher, the chance of a reduction is being reduced.
        """
        pieces: CommandArgList = [key, topk]
        if width is not None and depth is not None and decay is not None:
            pieces.extend([width, depth, decay])
        return await self.execute_module_command(
            CommandName.TOPK_RESERVE, *pieces, callback=SimpleStringCallback()
        )

    @module_command(
        CommandName.TOPK_ADD,
        group=COMMAND_GROUP,
        version_introduced="2.0.0",
        module=MODULE,
    )
    async def add(
        self, key: KeyT, items: Parameters[AnyStr]
    ) -> Tuple[Optional[AnyStr], ...]:
        """
        Increases the count of one or more items by increment

        :param key: Name of the TOP-K sketch.
        :param items: Item(s) to be added.

        """
        return await self.execute_module_command(
            CommandName.TOPK_ADD,
            key,
            *items,
            callback=TupleCallback[Optional[AnyStr]](),
        )

    @module_command(
        CommandName.TOPK_INCRBY,
        group=COMMAND_GROUP,
        version_introduced="2.0.0",
        module=MODULE,
    )
    async def incrby(
        self, key: KeyT, items: Mapping[AnyStr, int]
    ) -> Tuple[Optional[AnyStr], ...]:
        """
        Increases the count of one or more items by increment

        :param key: Name of the TOP-K sketch.
        :param items: Dictionary of items and their corresponding increment values.
        """
        return await self.execute_module_command(
            CommandName.TOPK_INCRBY,
            key,
            *dict_to_flat_list(items),
            callback=TupleCallback[Optional[AnyStr]](),
        )

    @module_command(
        CommandName.TOPK_QUERY,
        group=COMMAND_GROUP,
        version_introduced="2.0.0",
        module=MODULE,
        flags={CommandFlag.READONLY},
        cache_config=CacheConfig(lambda *a, **_: a[0]),
    )
    async def query(
        self,
        key: KeyT,
        items: Parameters[StringT],
    ) -> Tuple[bool, ...]:
        """
        Checks whether an item is one of Top-K items.
        Multiple items can be checked at once.

        :param key: Name of the TOP-K sketch.
        :param items: Item(s) to be queried.
        """
        pieces: CommandArgList = [key, *items]

        return await self.execute_module_command(
            CommandName.TOPK_QUERY, *pieces, callback=BoolsCallback()
        )

    @module_command(
        CommandName.TOPK_COUNT,
        group=COMMAND_GROUP,
        version_introduced="2.0.0",
        module=MODULE,
    )
    async def count(
        self,
        key: KeyT,
        items: Parameters[StringT],
    ) -> Tuple[int, ...]:
        """
        Return the count for one or more items are in a sketch

        :param key: The name of the TOP-K sketch.
        :param items: One or more items to count.
        """
        pieces: CommandArgList = [key, *items]

        return await self.execute_module_command(
            CommandName.TOPK_COUNT, *pieces, callback=TupleCallback[int]()
        )

    @module_command(
        CommandName.TOPK_LIST,
        group=COMMAND_GROUP,
        version_introduced="2.0.0",
        module=MODULE,
        flags={CommandFlag.READONLY},
        cache_config=CacheConfig(lambda *a, **_: a[0]),
    )
    async def list(
        self, key: KeyT, withcount: Optional[bool] = None
    ) -> Union[Dict[AnyStr, int], Tuple[AnyStr, ...]]:
        """
        Return full list of items in Top K list

        :param key: Name of the TOP-K sketch.
        :param withcount: Whether to include counts of each element.
        """
        pieces: CommandArgList = [key]
        if withcount:
            pieces.append(PureToken.WITHCOUNT)
            return await self.execute_module_command(
                CommandName.TOPK_LIST, *pieces, callback=DictCallback[AnyStr, int]()
            )
        else:
            return await self.execute_module_command(
                CommandName.TOPK_LIST, *pieces, callback=TupleCallback[AnyStr]()
            )

    @module_command(
        CommandName.TOPK_INFO,
        group=COMMAND_GROUP,
        version_introduced="2.0.0",
        module=MODULE,
        flags={CommandFlag.READONLY},
        cache_config=CacheConfig(lambda *a, **_: a[0]),
    )
    async def info(self, key: KeyT) -> Dict[AnyStr, int]:
        """
        Returns information about a sketch

        :param key: Name of the TOP-K sketch.
        :return: A dictionary containing the following information:
         - ``k``: The number of items tracked by the sketch.
         - ``width``: The width of the sketch.
         - ``depth``: The depth of the sketch.
         - ``decay``: The decay factor used by the sketch.
        """

        return await self.execute_module_command(
            CommandName.TOPK_INFO,
            key,
            callback=DictCallback[AnyStr, int](),
        )


@versionadded(version="4.12")
class TDigest(ModuleGroup[AnyStr]):
    MODULE = RedisBloom
    COMMAND_GROUP = CommandGroup.TDIGEST

    @module_command(
        CommandName.TDIGEST_CREATE,
        group=COMMAND_GROUP,
        version_introduced="2.4.0",
        module=MODULE,
    )
    async def create(self, key: KeyT, compression: Optional[int] = None) -> bool:
        """
        Allocates memory and initializes a new t-digest sketch

        :param key: The key name for the new t-digest sketch.
        :param compression: A controllable tradeoff between accuracy and memory consumption.
        """
        pieces: CommandArgList = [key]
        if compression is not None:
            pieces.extend([PrefixToken.COMPRESSION, compression])
        return await self.execute_module_command(
            CommandName.TDIGEST_CREATE, *pieces, callback=SimpleStringCallback()
        )

    @module_command(
        CommandName.TDIGEST_RESET,
        group=COMMAND_GROUP,
        version_introduced="2.4.0",
        module=MODULE,
    )
    async def reset(self, key: KeyT) -> bool:
        """
        Resets a t-digest sketch: empty the sketch and re-initializes it.

        :param key: The key name for an existing t-digest sketch.
        """
        return await self.execute_module_command(
            CommandName.TDIGEST_RESET, key, callback=SimpleStringCallback()
        )

    @module_command(
        CommandName.TDIGEST_ADD,
        group=COMMAND_GROUP,
        version_introduced="2.4.0",
        module=MODULE,
    )
    async def add(
        self,
        key: KeyT,
        values: Parameters[Union[int, float]],
    ) -> bool:
        """
        Adds one or more observations to a t-digest sketch

        :param key: Key name for an existing t-digest sketch.
        :param values: value(s) of observation(s)
        """
        pieces: CommandArgList = [key, *values]

        return await self.execute_module_command(
            CommandName.TDIGEST_ADD, *pieces, callback=SimpleStringCallback()
        )

    @module_command(
        CommandName.TDIGEST_MERGE,
        group=COMMAND_GROUP,
        version_introduced="2.4.0",
        module=MODULE,
    )
    async def merge(
        self,
        destination_key: KeyT,
        source_keys: Parameters[KeyT],
        compression: Optional[int] = None,
        override: Optional[bool] = None,
    ) -> bool:
        """
        Merges multiple t-digest sketches into a single sketch

        :param destination_key: Key name for a t-digest sketch to merge observation values to.
         If it does not exist, a new sketch is created.
         If it is an existing sketch, its values are merged with the values of the source keys.
         To override the destination key contents use :paramref:`override`.
        :param source_keys: Key names for t-digest sketches to merge observation values from.
        :param compression: Controllable tradeoff between accuracy and memory consumption.
        :param override: When specified, if :paramref:`destination_key` already exists,
         it is overwritten.

        """
        _source_keys: List[KeyT] = list(source_keys)
        pieces: CommandArgList = [
            destination_key,
            len(_source_keys),
            *_source_keys,
        ]
        if compression is not None:
            pieces.extend([PrefixToken.COMPRESSION, compression])
        if override is not None:
            pieces.append(PureToken.OVERRIDE)
        return await self.execute_module_command(
            CommandName.TDIGEST_MERGE, *pieces, callback=SimpleStringCallback()
        )

    @module_command(
        CommandName.TDIGEST_MIN,
        group=COMMAND_GROUP,
        version_introduced="2.4.0",
        module=MODULE,
        flags={CommandFlag.READONLY},
        cache_config=CacheConfig(lambda *a, **_: a[0]),
    )
    async def min(self, key: KeyT) -> float:
        """
        Returns the minimum observation value from a t-digest sketch

        :param key: The key name for an existing t-digest sketch.
        """

        return await self.execute_module_command(
            CommandName.TDIGEST_MIN, key, callback=FloatCallback()
        )

    @module_command(
        CommandName.TDIGEST_MAX,
        group=COMMAND_GROUP,
        version_introduced="2.4.0",
        module=MODULE,
        flags={CommandFlag.READONLY},
        cache_config=CacheConfig(lambda *a, **_: a[0]),
    )
    async def max(self, key: KeyT) -> float:
        """
        Returns the maximum observation value from a t-digest sketch

        :param key: The key name for an existing t-digest sketch.
        """

        return await self.execute_module_command(
            CommandName.TDIGEST_MAX, key, callback=FloatCallback()
        )

    @module_command(
        CommandName.TDIGEST_QUANTILE,
        group=COMMAND_GROUP,
        version_introduced="2.4.0",
        module=MODULE,
        flags={CommandFlag.READONLY},
        cache_config=CacheConfig(lambda *a, **_: a[0]),
    )
    async def quantile(
        self,
        key: KeyT,
        quantiles: Parameters[Union[int, float]],
    ) -> Tuple[float, ...]:
        """
        Returns, for each input fraction, an estimation of the value (floating point)
        that is smaller than the given fraction of observations

        :param key: Key name for an existing t-digest sketch.
        :param quantiles: Input fractions (between 0 and 1 inclusively).
        """
        pieces: CommandArgList = [key, *quantiles]

        return await self.execute_module_command(
            CommandName.TDIGEST_QUANTILE, *pieces, callback=FloatsCallback()
        )

    @module_command(
        CommandName.TDIGEST_CDF,
        group=COMMAND_GROUP,
        version_introduced="2.4.0",
        module=MODULE,
        flags={CommandFlag.READONLY},
        cache_config=CacheConfig(lambda *a, **_: a[0]),
    )
    async def cdf(
        self,
        key: KeyT,
        values: Parameters[Union[int, float]],
    ) -> Tuple[float, ...]:
        """
        Returns, for each input value, an estimation of the fraction (floating-point)
        of (observations smaller than the given value + half the observations equal
        to the given value)

        :param key: The key name for an existing t-digest sketch.
        :param values: The values for which the CDF should be retrieved.
        """
        pieces: CommandArgList = [key, *values]

        return await self.execute_module_command(
            CommandName.TDIGEST_CDF, *pieces, callback=FloatsCallback()
        )

    @module_command(
        CommandName.TDIGEST_TRIMMED_MEAN,
        group=COMMAND_GROUP,
        version_introduced="2.4.0",
        module=MODULE,
        flags={CommandFlag.READONLY},
        cache_config=CacheConfig(lambda *a, **_: a[0]),
    )
    async def trimmed_mean(
        self,
        key: KeyT,
        low_cut_quantile: Union[int, float],
        high_cut_quantile: Union[int, float],
    ) -> float:
        """
        Returns an estimation of the mean value from the sketch,
        excluding observation values outside the low and high cutoff quantiles

        :param key: The key name for an existing t-digest sketch.
        :param low_cut_quantile: A floating-point value in the range [0..1],
         should be lower than :paramref:`high_cut_quantile`.
        :param high_cut_quantile: A floating-point value in the range [0..1],
         should be higher than `low_cut_quantile`.

        """
        pieces: CommandArgList = [key, low_cut_quantile, high_cut_quantile]

        return await self.execute_module_command(
            CommandName.TDIGEST_TRIMMED_MEAN, *pieces, callback=FloatCallback()
        )

    @module_command(
        CommandName.TDIGEST_RANK,
        group=COMMAND_GROUP,
        version_introduced="2.4.0",
        module=MODULE,
        flags={CommandFlag.READONLY},
        cache_config=CacheConfig(lambda *a, **_: a[0]),
    )
    async def rank(
        self,
        key: KeyT,
        values: Parameters[Union[int, float]],
    ) -> Tuple[int, ...]:
        """
        Returns, for each input value (floating-point), the estimated rank of
        the value (the number of observations in the sketch that are smaller
        than the value + half the number of observations that are equal to the value)

        :param key: The key name for an existing t-digest sketch.
        :param values: Input values for which the rank should be estimated.
        """
        pieces: CommandArgList = [key, *values]

        return await self.execute_module_command(
            CommandName.TDIGEST_RANK, *pieces, callback=TupleCallback[int]()
        )

    @module_command(
        CommandName.TDIGEST_REVRANK,
        group=COMMAND_GROUP,
        version_introduced="2.4.0",
        module=MODULE,
        flags={CommandFlag.READONLY},
        cache_config=CacheConfig(lambda *a, **_: a[0]),
    )
    async def revrank(
        self,
        key: KeyT,
        values: Parameters[Union[int, float]],
    ) -> Tuple[int, ...]:
        """
        Returns, for each input value (floating-point), the estimated reverse rank of
        the value (the number of observations in the sketch that are larger than
        the value + half the number of observations that are equal to the value)

        :param key: The name of an existing t-digest sketch.
        :param values: The input values for which the reverse rank should be estimated.
        """
        pieces: CommandArgList = [key, *values]

        return await self.execute_module_command(
            CommandName.TDIGEST_REVRANK, *pieces, callback=TupleCallback[int]()
        )

    @module_command(
        CommandName.TDIGEST_BYRANK,
        group=COMMAND_GROUP,
        version_introduced="2.4.0",
        module=MODULE,
        flags={CommandFlag.READONLY},
        cache_config=CacheConfig(lambda *a, **_: a[0]),
    )
    async def byrank(
        self,
        key: KeyT,
        ranks: Parameters[Union[int, float]],
    ) -> Tuple[float, ...]:
        """
        Returns, for each input rank, an estimation of the value (floating-point) with
        that rank

        :param key: The key name for an existing t-digest sketch.
        :param ranks: The ranks for which the estimated values should be retrieved.
        """
        pieces: CommandArgList = [key, *ranks]

        return await self.execute_module_command(
            CommandName.TDIGEST_BYRANK, *pieces, callback=FloatsCallback()
        )

    @module_command(
        CommandName.TDIGEST_BYREVRANK,
        group=COMMAND_GROUP,
        version_introduced="2.4.0",
        module=MODULE,
        flags={CommandFlag.READONLY},
        cache_config=CacheConfig(lambda *a, **_: a[0]),
    )
    async def byrevrank(
        self,
        key: KeyT,
        reverse_ranks: Parameters[Union[int, float]],
    ) -> Tuple[float, ...]:
        """
        Returns, for each input reverse rank, an estimation of the value
        (floating-point) with that reverse rank

        :param key: The key name for an existing t-digest sketch.
        :param reverse_ranks: The reverse ranks for which the values should be retrieved.
        """
        pieces: CommandArgList = [key, *reverse_ranks]

        return await self.execute_module_command(
            CommandName.TDIGEST_BYREVRANK, *pieces, callback=FloatsCallback()
        )

    @module_command(
        CommandName.TDIGEST_INFO,
        group=COMMAND_GROUP,
        version_introduced="2.4.0",
        module=MODULE,
        flags={CommandFlag.READONLY},
        cache_config=CacheConfig(lambda *a, **_: a[0]),
    )
    async def info(self, key: KeyT) -> Dict[AnyStr, ResponsePrimitive]:
        """
        Returns information and statistics about a t-digest sketch

        :param key: The key name for an existing t-digest sketch.
        :return: Dictionary with information about the sketch, including compression,
         capacity, number of merged and unmerged nodes, weight of merged and unmerged nodes,
         number of observations, total compressions, and memory usage.
        """

        return await self.execute_module_command(
            CommandName.TDIGEST_INFO,
            key,
            callback=DictCallback[AnyStr, ResponsePrimitive](),
        )
